from abc import abstractmethod
from typing import Optional, Dict

from deepeval.test_case import LLMTestCase, ConversationalTestCase


class BaseMetric:
    score: Optional[float] = None
    score_breakdown: Dict = None
    reason: Optional[str] = None
    success: Optional[bool] = None
    evaluation_model: Optional[str] = None
    strict_mode: bool = False
    async_mode: bool = True
    verbose_mode: bool = True
    include_reason: bool = False
    error: Optional[str] = None
    evaluation_cost: Optional[float] = None
    verbose_logs: Optional[str] = None

    @property
    def threshold(self) -> float:
        return self._threshold

    @threshold.setter
    def threshold(self, value: float):
        self._threshold = value

    @abstractmethod
    def measure(self, test_case: LLMTestCase, *args, **kwargs) -> float:
        raise NotImplementedError

    @abstractmethod
    async def a_measure(self, test_case: LLMTestCase, *args, **kwargs) -> float:
        raise NotImplementedError(
            f"Async execution for {self.__class__.__name__} not supported yet. Please set 'async_mode' to 'False'."
        )

    @abstractmethod
    def is_successful(self) -> bool:
        raise NotImplementedError

    @property
    def __name__(self):
        return "Base Metric"


class BaseConversationalMetric:
    score: Optional[float] = None
    score_breakdown: Dict = None
    reason: Optional[str] = None
    evaluation_model: Optional[str] = None
    error: Optional[str] = None
    # Not changeable for now
    strict_mode: bool = False
    async_mode: bool = False

    @property
    def threshold(self) -> float:
        return self._threshold

    @threshold.setter
    def threshold(self, value: float):
        self._threshold = value

    @abstractmethod
    def measure(
        self, test_case: ConversationalTestCase, *args, **kwargs
    ) -> float:
        raise NotImplementedError

    @abstractmethod
    def is_successful(self) -> bool:
        raise NotImplementedError

    @property
    def __name__(self):
        return "Base Conversational Metric"



from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    Iterable,
    Iterator,
    Tuple,
    Mapping,
    cast,
)
from time import perf_counter
import json
import traceback

from llama_index.core.bridge.pydantic import BaseModel
from llama_index.core.callbacks.base_handler import BaseCallbackHandler
from llama_index.core.callbacks.schema import CBEventType, EventPayload
from llama_index.core.llms import ChatMessage
from llama_index.core.schema import BaseNode
from llama_index.core import Response
from llama_index.core.base.response.schema import StreamingResponse
from llama_index.core.callbacks import CBEventType, EventPayload
from llama_index.core.callbacks.base_handler import BaseCallbackHandler
from llama_index.core.callbacks.schema import BASE_TRACE_EVENT
from llama_index.core.llms import ChatMessage, ChatResponse
from llama_index.core.tools import ToolMetadata

from deepeval.tracing import (
    trace_manager,
    get_trace_stack,
    BaseTrace,
    LlmTrace,
    GenericTrace,
    EmbeddingTrace,
    RerankingTrace,
    RetrieverTrace,
    TraceStatus,
    LlmAttributes,
    EmbeddingAttributes,
    RerankingAttributes,
    RetrieverAttributes,
    TraceType,
    TraceProvider,
    LlamaIndexTraceType,
    RetrievalNode,
    QueryTrace,
    QueryAttributes,
    SynthesizeAttributes,
    SynthesizeTrace,
    GenericAttributes,
    AgentAttributes,
    AgentTrace,
    TraceData,
)
from deepeval.utils import dataclass_to_dict
from deepeval.event import track

events_to_ignore = [
    # CBEventType.CHUNKING,
    # CBEventType.NODE_PARSING,
    # CBEventType.EMBEDDING,
    # CBEventType.LLM,
    # CBEventType.QUERY,
    # CBEventType.RETRIEVE,
    # CBEventType.SYNTHESIZE,
    CBEventType.TREE,
    CBEventType.SUB_QUESTION,
    # CBEventType.TEMPLATING,
    CBEventType.FUNCTION_CALL,
    # CBEventType.RERANKING,
    # CBEventType.EXCEPTION,
    # CBEventType.AGENT_STEP,
]


class LlamaIndexCallbackHandler(BaseCallbackHandler):
    def __init__(self) -> None:
        self.track_params = {}
        self.event_map = {}
        self._templating_parent_id = {}
        self._templating_payloads = {}
        super().__init__(
            event_starts_to_ignore=events_to_ignore,
            event_ends_to_ignore=events_to_ignore,
        )

    def start_trace(self, trace_id: Optional[str] = None) -> None:
        self.event_map = {}
        # trace_manager.clear_trace_stack()
        return

    def end_trace(
        self,
        trace_id: Optional[str] = None,
        trace_map: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        return

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> str:
        # set outtermost provider
        if not trace_manager.get_outter_provider():
            trace_manager.set_outter_provider(TraceProvider.LLAMA_INDEX)

        processed_payload = self.process_payload(
            event_type, event_id, parent_id, payload, True
        )
        trace_instance = self.create_trace_instance(
            event_type, processed_payload
        )
        self.event_map[event_id] = trace_instance
        trace_manager.append_to_trace_stack(trace_instance)

        return

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> None:

        trace_instance = self.event_map[event_id]
        processed_payload = self.process_payload(
            event_type, event_id, parent_id, payload, True
        )
        trace_instance = self.update_trace_instance(
            trace_instance, event_type, processed_payload
        )
        current_trace_stack = trace_manager.get_trace_stack_copy()
        # trace_instance = current_trace_stack[-1]

        if len(current_trace_stack) > 1:
            parent_trace = current_trace_stack[-2]
            parent_trace.traces.append(trace_instance)
            trace_manager.set_trace_stack(current_trace_stack)

        if len(current_trace_stack) == 1:
            dict_representation = dataclass_to_dict(current_trace_stack[0])
            trace_manager.set_dict_trace_stack(dict_representation)
            trace_manager.clear_trace_stack()

            if trace_manager.get_outter_provider() == TraceProvider.LLAMA_INDEX:
                track(
                    event_name=current_trace_stack[0].name,
                    model=self.track_params.get("model") or "NA",
                    input=self.track_params.get("input") or "NA",
                    response=self.track_params.get("response") or "NA",
                    retrieval_context=self.track_params.get(
                        "retrieval_context"
                    ),
                    completion_time=current_trace_stack[0].executionTime,
                    token_usage=self.track_params.get("token_usage"),
                    trace_stack=dict_representation,
                )
        else:
            trace_manager.pop_trace_stack()

        return

    def create_trace_instance(
        self,
        event_type: CBEventType,
        processed_payload: Optional[Dict[str, Any]] = None,
    ) -> TraceData:

        trace_kwargs = {
            "traceProvider": TraceProvider.LLAMA_INDEX,
            "type": self.convert_event_type_to_deepeval_trace_type(event_type),
            "executionTime": perf_counter(),
            "name": event_type,
            "status": TraceStatus.SUCCESS,
            "traces": [],
        }

        if "exception" in processed_payload:
            trace_kwargs["status"] = TraceStatus.ERROR
            trace_instance = GenericTrace(
                **trace_kwargs,
            )

        ### Different Attributes ###############################

        elif event_type == CBEventType.AGENT_STEP:
            trace_instance = AgentTrace(
                **trace_kwargs,
                embeddingAttributes=AgentAttributes(
                    input="", output="", name="", description=""
                ),
            )

        elif event_type == CBEventType.EMBEDDING:
            trace_instance = EmbeddingTrace(
                **trace_kwargs,
                embeddingAttributes=EmbeddingAttributes(
                    embedding_text="",
                    # Optional variables
                    model=processed_payload["embedding_model_name"],
                    embedding_length=None,
                ),
            )

        elif event_type == CBEventType.LLM:
            messages = processed_payload.get("llm_input_messages")
            trace_instance = LlmTrace(
                **trace_kwargs,
                llmAttributes=LlmAttributes(
                    input_str=next(
                        m["message_content"]
                        for m in messages
                        if m["message_role"] == "user"
                    ),
                    output_str="",
                    # Optional variables
                    model=processed_payload["llm_model_name"],
                    total_token_count=None,
                    prompt_token_count=None,
                    completion_token_count=None,
                    prompt_template=processed_payload.get(
                        "llm_prompt_template"
                    ),
                    prompt_template_variables=processed_payload.get(
                        "llm_prompt_template_variables"
                    ),
                ),
            )
            self.track_params["model"] = processed_payload["llm_model_name"]

        elif event_type == CBEventType.RERANKING:
            trace_instance = RerankingTrace(
                **trace_kwargs,
                rerankingAttributes=RerankingAttributes(
                    input_nodes=[],
                    output_nodes=[],
                    # Optional variables
                    model=processed_payload["reranker_model_name"],
                    top_n=processed_payload["reranker_top_k"],
                    batch_size=None,
                    query_str=None,
                ),
            )

        elif event_type == CBEventType.RETRIEVE:
            trace_instance = RetrieverTrace(
                **trace_kwargs,
                retrieverAttributes=RetrieverAttributes(
                    query_str=processed_payload["input_value"],
                    nodes=[],
                    # Optional variables
                    top_k=None,
                    average_chunk_size=None,
                    top_score=None,
                    similarity_scorer=None,
                ),
            )

        elif event_type == CBEventType.QUERY:
            trace_instance = QueryTrace(
                **trace_kwargs,
                queryAttributes=QueryAttributes(
                    input=processed_payload["input_value"], output=""
                ),
            )
            self.track_params["input"] = processed_payload["input_value"]

        elif event_type == CBEventType.SYNTHESIZE:
            trace_instance = SynthesizeTrace(
                **trace_kwargs,
                synthesizeAttributes=SynthesizeAttributes(
                    user_query=processed_payload["input_value"],
                    response="",
                    # Optional variables
                    retrieved_context=None,
                ),
            )

        else:
            trace_instance = GenericTrace(
                **trace_kwargs, genericAttributes=None
            )

        return trace_instance

    def update_trace_instance(
        self,
        trace_instance: BaseTrace,
        event_type: CBEventType,
        processed_payload: Optional[Dict[str, Any]] = None,
    ) -> TraceData:

        trace_instance.executionTime = (
            perf_counter() - trace_instance.executionTime
        )

        if "exception" in processed_payload:
            trace_instance.status = TraceStatus.ERROR

        elif event_type == CBEventType.LLM and isinstance(
            trace_instance, LlmTrace
        ):
            attributes = trace_instance.llmAttributes
            attributes.output_str = processed_payload["output_value"]
            attributes.total_token_count = processed_payload[
                "llm_token_count_total"
            ]
            attributes.prompt_token_count = processed_payload[
                "llm_token_prompt_count"
            ]
            attributes.completion_token_count = processed_payload[
                "llm_token_count_completion"
            ]
            self.track_params["token_usage"] = processed_payload[
                "llm_token_count_total"
            ]

        elif event_type == CBEventType.EMBEDDING and isinstance(
            trace_instance, EmbeddingTrace
        ):
            attributes = trace_instance.embeddingAttributes
            embedding = processed_payload["embeddings"][0]
            attributes.embedding_text = embedding["embedding_text"]
            attributes.embedding_length = len(embedding["embedding_vector"])

        elif event_type == CBEventType.RETRIEVE and isinstance(
            trace_instance, RetrieverTrace
        ):
            attributes = trace_instance.retrieverAttributes
            total_chunk_length = 0
            top_score = 0
            nodes: List[RetrievalNode] = processed_payload[
                "retrieval_documents"
            ]
            for node in nodes:
                total_chunk_length += len(node.content)
                top_score = node.score if node.score > top_score else top_score
            attributes.nodes = nodes
            attributes.top_k = len(nodes)
            attributes.average_chunk_size = total_chunk_length // len(nodes)
            attributes.top_score = top_score
            self.track_params["retrieval_context"] = [
                node.content for node in nodes
            ]

        elif event_type == CBEventType.QUERY and isinstance(
            trace_instance, QueryTrace
        ):
            attributes = trace_instance.queryAttributes
            attributes.output = processed_payload["output_value"]
            self.track_params["response"] = processed_payload["output_value"]

        elif event_type == CBEventType.SYNTHESIZE and isinstance(
            trace_instance, SynthesizeTrace
        ):
            attributes = trace_instance.synthesizeAttributes
            attributes.response = processed_payload["output_value"]

        return trace_instance

    def convert_event_type_to_deepeval_trace_type(
        self, event_type: CBEventType
    ):
        # TODO: add more types
        if event_type == CBEventType.LLM:
            return LlamaIndexTraceType.LLM
        elif event_type == CBEventType.RETRIEVE:
            return LlamaIndexTraceType.RETRIEVER
        elif event_type == CBEventType.EMBEDDING:
            return LlamaIndexTraceType.EMBEDDING
        elif event_type == CBEventType.SYNTHESIZE:
            return LlamaIndexTraceType.SYNTHESIZE
        elif event_type == CBEventType.QUERY:
            return LlamaIndexTraceType.QUERY
        elif event_type == CBEventType.RERANKING:
            return LlamaIndexTraceType.RERANKING
        elif event_type == CBEventType.AGENT_STEP:
            return LlamaIndexTraceType.AGENT

        return event_type.value.capitalize()

    def process_payload(
        self,
        event_type: CBEventType,
        event_id: str,
        parent_id: str,
        payload: Optional[Dict[str, Any]] = None,
        is_event_end: bool = True,
    ):
        attributes = {}
        if payload == None:
            return attributes

        exception = payload.get(EventPayload.EXCEPTION)
        if exception:
            attributes["exception"] = exception

        #########################
        # ignore these events
        if event_type in (CBEventType.NODE_PARSING, CBEventType.CHUNKING):
            return attributes

        #########################
        # process embedding vectors
        chunks = payload.get(EventPayload.CHUNKS)
        embeddings = payload.get(EventPayload.EMBEDDINGS)
        if chunks and embeddings:
            attributes["embeddings"] = [
                {"embedding_text": text, "embedding_vector": vector}
                for text, vector in zip(chunks, embeddings)
            ]

        #########################
        # process retrieval documents
        nodes = payload.get(EventPayload.NODES)
        if event_type is not CBEventType.RERANKING and nodes:
            attributes["retrieval_documents"] = self.process_nodes(nodes)

        #########################
        # process input (query string, prompts, messages)
        query_str = payload.get(EventPayload.QUERY_STR)
        prompt = payload.get(EventPayload.PROMPT)
        messages = payload.get(EventPayload.MESSAGES)

        if event_type is not CBEventType.RERANKING and query_str:
            attributes["input_value"] = query_str
        if prompt:
            attributes["llm_prompts"] = [prompt]
        if messages:
            if event_type is CBEventType.LLM:
                llm_msgs = [self.process_message(m) for m in messages]
                attributes["llm_input_messages"] = llm_msgs
            elif event_type is CBEventType.AGENT_STEP and len(messages):
                msg = messages[0]
                str_msg = (
                    msg.content if isinstance(msg, ChatMessage) else str(msg)
                )
                attributes["input_value"] = str_msg

        #########################
        # process response (still need to process token)
        response = payload.get(EventPayload.RESPONSE) or payload.get(
            EventPayload.COMPLETION
        )

        if response:
            attributes.update(self._get_response_output(response))
            if hasattr(response, "raw"):
                raw = response.raw
                assert isinstance(
                    raw, Mapping
                ), f"raw must be Mapping, found {type(raw)}"
                attributes.update(self._get_output_messages(raw))
                if "usage" in raw:
                    attributes.update(self._get_token_counts(raw["usage"]))

            if (
                additional_kwargs := getattr(
                    response, "additional_kwargs", None
                )
            ) is not None:
                attributes.update(self._get_token_counts(additional_kwargs))

        ###########################
        # process reranking
        reranker_top_k = payload.get(EventPayload.TOP_K)
        reranker_model_name = payload.get(EventPayload.MODEL_NAME)
        query_str = payload.get(EventPayload.QUERY_STR)
        nodes = payload.get(EventPayload.NODES)

        if event_type is CBEventType.RERANKING:
            if reranker_top_k:
                attributes["reranker_top_k"] = reranker_top_k
            if reranker_model_name:
                attributes["reranker_model_name"] = reranker_model_name
            if query_str:
                attributes["reranker_query"] = query_str
            if nodes:
                processed_nodes = self.process_nodes(nodes)
                if is_event_end:
                    attributes["reranker_output_documents"] = processed_nodes
                else:
                    attributes["reranker_input_documents"] = processed_nodes

        ###########################
        # process tool
        tool = payload.get(EventPayload.TOOL)
        tool_parameters = None

        if tool:
            tool_metadata = cast(ToolMetadata, tool)
            tool_parameters = tool_metadata.to_openai_tool()["function"][
                "parameters"
            ]
            attributes["tool_name"] = tool_metadata.name
            attributes["tool_description"] = tool_metadata.description

        if tool_parameters:
            attributes["tool_parameters"] = json.dumps(tool_parameters)

        ###########################
        # process tool
        serialized = payload.get(EventPayload.SERIALIZED)
        if serialized:
            if event_type is CBEventType.EMBEDDING:
                if model_name := serialized.get("model_name"):
                    attributes["embedding_model_name"] = model_name
            if event_type is CBEventType.LLM:
                if model_name := serialized.get("model"):
                    attributes["llm_model_name"] = model_name
                    invocation_parameters = self._extract_invocation_parameters(
                        serialized
                    )
                    invocation_parameters["model"] = model_name
                    attributes["llm_hyperparameters"] = json.dumps(
                        invocation_parameters
                    )

        ###########################
        # process templates
        payloads = [payload.copy()]
        if is_event_end == False:
            if event_type is CBEventType.TEMPLATING:
                self._templating_parent_id[event_id] = parent_id
                if payloads:
                    if parent_id in self._templating_payloads:
                        self._templating_payloads[parent_id].extend(payloads)
                    else:
                        self._templating_payloads[parent_id] = payloads

            if event_type is CBEventType.LLM:
                for templating_payload in self._templating_payloads.pop(
                    parent_id, ()
                ):
                    attributes.update(
                        self._template_attributes(templating_payload)
                    )
        else:
            if event_type is CBEventType.TEMPLATING:
                if (
                    parent_id := self._templating_parent_id.pop(event_id, None)
                ) and payload:
                    if parent_id in self._templating_payloads:
                        self._templating_payloads[parent_id].append(payload)
                    else:
                        self._templating_payloads[parent_id] = [payload]

        return attributes

    ##################################################
    ########### additional helper functions ##########
    ##################################################

    def _get_response_output(self, response: Any) -> Iterator[Tuple[str, Any]]:
        if isinstance(response, ChatResponse):
            message = response.message
            if content := message.content:
                yield "output_value", content
            else:
                yield "output_value", json.dumps(
                    message.additional_kwargs, cls=_CustomJSONEncoder
                )
        elif isinstance(response, Response):
            if response.response:
                yield "output_value", response.response
        elif isinstance(response, StreamingResponse):
            if response_txt := getattr(response, "response_txt", None):
                yield "output_value", response_txt
        else:  # if the response has unknown type, make a best-effort attempt to get the output
            yield "output_value", str(response)

    def _get_token_counts(
        self, usage: Union[object, Mapping[str, Any]]
    ) -> Iterator[Tuple[str, Any]]:
        if isinstance(usage, Mapping):
            yield from self._get_token_counts_from_mapping(usage)
        elif isinstance(usage, object):
            yield from self._get_token_counts_from_object(usage)

    def _template_attributes(
        self, payload: Dict[str, Any]
    ) -> Iterator[Tuple[str, Any]]:
        if template := payload.get(EventPayload.TEMPLATE):
            yield "llm_prompt_template", template
        if template_vars := payload.get(EventPayload.TEMPLATE_VARS):
            yield "llm_prompt_template_variables", json.dumps(template_vars)

    def _get_token_counts_from_object(
        self, usage: object
    ) -> Iterator[Tuple[str, Any]]:
        if (prompt_tokens := getattr(usage, "prompt_tokens", None)) is not None:
            yield "llm_token_prompt_count", prompt_tokens
        if (
            completion_tokens := getattr(usage, "completion_tokens", None)
        ) is not None:
            yield "llm_token_count_completion", completion_tokens
        if (total_tokens := getattr(usage, "total_tokens", None)) is not None:
            yield "llm_token_count_total", total_tokens

    def _get_token_counts_from_mapping(
        self,
        usage_mapping: Mapping[str, Any],
    ) -> Iterator[Tuple[str, Any]]:
        if (prompt_tokens := usage_mapping.get("prompt_tokens")) is not None:
            yield "llm_token_prompt_count", prompt_tokens
        if (
            completion_tokens := usage_mapping.get("completion_tokens")
        ) is not None:
            yield "llm_token_count_completion", completion_tokens
        if (total_tokens := usage_mapping.get("total_tokens")) is not None:
            yield "llm_token_count_total", total_tokens

    def process_nodes(self, nodes) -> Dict[str, Optional[str]]:
        processed_nodes = [
            RetrievalNode(
                content=node_with_score.node.text,
                id=node_with_score.node.node_id,
                score=node_with_score.score,
            )
            for node_with_score in nodes
        ]
        return processed_nodes

    def process_message(self, message: Any) -> Dict[str, Optional[str]]:
        if isinstance(message, ChatMessage):
            message_attributes = {
                "message_role": message.role.value,
                "message_content": message.content,
            }
            # Parse the kwargs to extract the function name and parameters for function calling
            # NB: these additional kwargs exist both for 'agent' and 'function' roles
            if "name" in message.additional_kwargs:
                message_attributes["message_name"] = message.additional_kwargs[
                    "name"
                ]
            if tool_calls := message.additional_kwargs.get("tool_calls"):
                assert isinstance(
                    tool_calls, Iterable
                ), f"tool_calls must be Iterable, found {type(tool_calls)}"
                message_tool_calls = []
                for tool_call in tool_calls:
                    if message_tool_call := dict(
                        self._get_tool_call(tool_call)
                    ):
                        message_tool_calls.append(message_tool_call)
                if message_tool_calls:
                    message_attributes["message_tool_calls"] = (
                        message_tool_calls
                    )
            return message_attributes

        return {
            "message_role": "user",  # assume user if not ChatMessage
            "message_content": str(message),
        }

    def _get_tool_call(self, tool_call: object) -> Iterator[Tuple[str, Any]]:
        if function := getattr(tool_call, "function", None):
            if name := getattr(function, "name", None):
                assert isinstance(
                    name, str
                ), f"name must be str, found {type(name)}"
                yield "tool_call_function_name", name
            if arguments := getattr(function, "arguments", None):
                assert isinstance(
                    arguments, str
                ), f"arguments must be str, found {type(arguments)}"
                yield "tool_call_function_arguments_json", arguments

    def _extract_invocation_parameters(
        self, serialized: Mapping[str, Any]
    ) -> Dict[str, Any]:
        # FIXME: this is only based on openai. Other models have different parameters.
        if not hasattr(serialized, "get"):
            return {}
        invocation_parameters: Dict[str, Any] = {}
        additional_kwargs = serialized.get("additional_kwargs")
        if additional_kwargs and isinstance(additional_kwargs, Mapping):
            invocation_parameters.update(additional_kwargs)
        for key in ("temperature", "max_tokens"):
            if (value := serialized.get(key)) is not None:
                invocation_parameters[key] = value
        return invocation_parameters

    def _get_output_messages(
        self, raw: Mapping[str, Any]
    ) -> Iterator[Tuple[str, Any]]:
        assert hasattr(raw, "get"), f"raw must be Mapping, found {type(raw)}"
        if not (choices := raw.get("choices")):
            return
        assert isinstance(
            choices, Iterable
        ), f"choices must be Iterable, found {type(choices)}"
        if messages := [
            dict(self._get_message(message))
            for choice in choices
            if (message := getattr(choice, "message", None)) is not None
        ]:
            yield "llm_output_messages", messages

    def _get_message(self, message: object) -> Iterator[Tuple[str, Any]]:
        if role := getattr(message, "role", None):
            assert isinstance(
                role, str
            ), f"content must be str, found {type(role)}"
            yield "message_role", role
        if content := getattr(message, "content", None):
            assert isinstance(
                content, str
            ), f"content must be str, found {type(content)}"
            yield "message_content", content
        if tool_calls := getattr(message, "tool_calls", None):
            assert isinstance(
                tool_calls, Iterable
            ), f"tool_calls must be Iterable, found {type(tool_calls)}"
            message_tool_calls = []
            for tool_call in tool_calls:
                if message_tool_call := dict(self._get_tool_call(tool_call)):
                    message_tool_calls.append(message_tool_call)
            if message_tool_calls:
                yield "message_tool_calls", message_tool_calls


class _CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        try:
            return super().default(obj)
        except TypeError:
            if callable(as_dict := getattr(obj, "dict", None)):
                return as_dict()
            raise



import asyncio
from pydantic import BaseModel

import logging
import openai
import instructor
from openai import OpenAI, AsyncOpenAI
from pydantic import BaseModel

from typing import Optional, Tuple
from langchain_openai import ChatOpenAI, AzureChatOpenAI
from langchain_community.callbacks import get_openai_callback
from tenacity import retry, retry_if_exception_type, wait_exponential_jitter

from deepeval.key_handler import KeyValues, KEY_FILE_HANDLER
from deepeval.models import DeepEvalBaseLLM


def log_retry_error(retry_state):
    logging.error(
        f"OpenAI rate limit exceeded. Retrying: {retry_state.attempt_number} time(s)..."
    )


valid_gpt_models = [
    "gpt-4o-mini",
    "gpt-4o",
    "gpt-4-turbo",
    "gpt-4-turbo-preview",
    "gpt-4-0125-preview",
    "gpt-4-1106-preview",
    "gpt-4",
    "gpt-4-32k",
    "gpt-4-0613",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0125",
]

default_gpt_model = "gpt-4o"

#########################################################
##### custom model with pydantic_model argument
#########################################################


class CustomGPTEnforced(DeepEvalBaseLLM):
    def __init__(
        self,
        model: Optional[str] = None,
        _openai_api_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        model_name = None
        if isinstance(model, str):
            model_name = model
            if model_name not in valid_gpt_models:
                raise ValueError(
                    f"Invalid model. Available GPT models: {', '.join(model for model in valid_gpt_models)}"
                )
        elif model is None:
            model_name = default_gpt_model

        self._openai_api_key = _openai_api_key
        self.is_azure_model: bool
        # args and kwargs will be passed to the underlying model, in load_model function
        self.args = args
        self.kwargs = kwargs
        super().__init__(model_name)

    def load_model(self):
        if self.should_use_azure_openai():
            openai_api_key = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_API_KEY
            )

            openai_api_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.OPENAI_API_VERSION
            )
            azure_deployment = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_DEPLOYMENT_NAME
            )
            azure_endpoint = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_ENDPOINT
            )

            model_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_MODEL_VERSION
            )

            if model_version is None:
                model_version = ""

            return AzureChatOpenAI(
                openai_api_version=openai_api_version,
                azure_deployment=azure_deployment,
                azure_endpoint=azure_endpoint,
                openai_api_key=openai_api_key,
                model_version=model_version,
                *self.args,
                **self.kwargs,
            )

        return ChatOpenAI(
            model_name=self.model_name,
            openai_api_key=self._openai_api_key,
            *self.args,
            **self.kwargs,
        )

    @retry(
        wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
        retry=retry_if_exception_type(openai.RateLimitError),
        after=log_retry_error,
    )
    def generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[str, float]:
        client = instructor.from_openai(OpenAI())
        return client.chat.completions.create(
            model=self.model_name,
            response_model=schema,
            messages=[{"role": "user", "content": prompt}],
        )

    @retry(
        wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
        retry=retry_if_exception_type(openai.RateLimitError),
        after=log_retry_error,
    )
    async def a_generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[str, float]:
        client = instructor.from_openai(AsyncOpenAI())
        response = await client.chat.completions.create(
            model=self.model_name,
            response_model=schema,
            messages=[{"role": "user", "content": prompt}],
        )
        return response

    def should_use_azure_openai(self):
        value = KEY_FILE_HANDLER.fetch_data(KeyValues.USE_AZURE_OPENAI)
        return value.lower() == "yes" if value is not None else False

    def get_model_name(self):
        if self.should_use_azure_openai():
            return "azure openai"
        elif self.model_name:
            return self.model_name


#########################################################
##### custom model with no pydantic_model argument
#########################################################


class CustomGPT(DeepEvalBaseLLM):
    def __init__(
        self,
        model: Optional[str] = None,
        _openai_api_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        model_name = None
        if isinstance(model, str):
            model_name = model
            if model_name not in valid_gpt_models:
                raise ValueError(
                    f"Invalid model. Available GPT models: {', '.join(model for model in valid_gpt_models)}"
                )
        elif model is None:
            model_name = default_gpt_model

        self._openai_api_key = _openai_api_key
        self.is_azure_model: bool
        # args and kwargs will be passed to the underlying model, in load_model function
        self.args = args
        self.kwargs = kwargs
        super().__init__(model_name)

    def load_model(self):
        if self.should_use_azure_openai():
            openai_api_key = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_API_KEY
            )

            openai_api_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.OPENAI_API_VERSION
            )
            azure_deployment = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_DEPLOYMENT_NAME
            )
            azure_endpoint = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_ENDPOINT
            )

            model_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_MODEL_VERSION
            )

            if model_version is None:
                model_version = ""

            return AzureChatOpenAI(
                openai_api_version=openai_api_version,
                azure_deployment=azure_deployment,
                azure_endpoint=azure_endpoint,
                openai_api_key=openai_api_key,
                model_version=model_version,
                *self.args,
                **self.kwargs,
            )

        return ChatOpenAI(
            model_name=self.model_name,
            openai_api_key=self._openai_api_key,
            *self.args,
            **self.kwargs,
        )

    @retry(
        wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
        retry=retry_if_exception_type(openai.RateLimitError),
        after=log_retry_error,
    )
    def generate(
        self,
        prompt: str,
    ) -> Tuple[str, float]:
        chat_model = self.load_model()
        with get_openai_callback() as cb:
            res = chat_model.invoke(prompt)
            return res.content

    @retry(
        wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
        retry=retry_if_exception_type(openai.RateLimitError),
        after=log_retry_error,
    )
    async def a_generate(
        self,
        prompt: str,
    ) -> Tuple[str, float]:
        chat_model = self.load_model()
        with get_openai_callback() as cb:
            res = await chat_model.ainvoke(prompt)
            return res.content

    @retry(
        wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=2, max=10),
        retry=retry_if_exception_type(openai.RateLimitError),
        after=log_retry_error,
    )
    def should_use_azure_openai(self):
        value = KEY_FILE_HANDLER.fetch_data(KeyValues.USE_AZURE_OPENAI)
        return value.lower() == "yes" if value is not None else False

    def get_model_name(self):
        if self.should_use_azure_openai():
            return "azure openai"
        elif self.model_name:
            return self.model_name


#########################################################
##### Define Metrics
#########################################################

from deepeval.metrics import (
    AnswerRelevancyMetric,
    BiasMetric,
    ContextualPrecisionMetric,
    ContextualRecallMetric,
    ContextualRelevancyMetric,
    FaithfulnessMetric,
    GEval,
    HallucinationMetric,
    SummarizationMetric,
    ToxicityMetric,
)
from deepeval.test_case import LLMTestCaseParams
from deepeval.test_case import LLMTestCase

test_case = LLMTestCase(
    input="What is this",
    actual_output="this is a latte",
    expected_output="this is a mocha",
    retrieval_context=["I love coffee"],
    context=["I love coffee"],
)

answer_relevancy = AnswerRelevancyMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
answer_relevancy_non_confine = AnswerRelevancyMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
answer_relevancy_confine = AnswerRelevancyMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

bias = BiasMetric(verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125")
bias_non_confine = BiasMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
bias_confine = BiasMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

contextual_precision = ContextualPrecisionMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
contextual_precision_non_confine = ContextualPrecisionMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
contextual_precision_confine = ContextualPrecisionMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

contextual_recall = ContextualRecallMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
contextual_recall_non_confine = ContextualRecallMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
contextual_recall_confine = ContextualRecallMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

contextual_relevancy = ContextualRelevancyMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
contextual_relevancy_non_confine = ContextualRelevancyMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
contextual_relevancy_confine = ContextualRelevancyMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

faithfulness = FaithfulnessMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
faithfulness_non_confine = FaithfulnessMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
faithfulness_confine = FaithfulnessMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

hallucination = HallucinationMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
hallucination_non_confine = HallucinationMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
hallucination_confine = HallucinationMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

summarization = SummarizationMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
summarization_non_confine = SummarizationMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
summarization_confine = SummarizationMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

toxicity = ToxicityMetric(
    verbose_mode=True, threshold=0.5, model="gpt-3.5-turbo-0125"
)
toxicity_non_confine = ToxicityMetric(
    verbose_mode=True, threshold=0.5, model=CustomGPT("gpt-3.5-turbo-0125")
)
toxicity_confine = ToxicityMetric(
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

g_eval = GEval(
    name="coherence",
    evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT],
    evaluation_steps=[
        "Response should be concise",
        "Reponse should read easily",
    ],
    verbose_mode=True,
    threshold=0.5,
    model="gpt-3.5-turbo-0125",
)
g_eval_non_confine = GEval(
    name="coherence",
    evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT],
    evaluation_steps=[
        "Response should be concise",
        "Reponse should read easily",
    ],
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPT("gpt-3.5-turbo-0125"),
)
g_eval_confine = GEval(
    name="coherence",
    evaluation_params=[LLMTestCaseParams.ACTUAL_OUTPUT],
    evaluation_steps=[
        "Response should be concise",
        "Reponse should read easily",
    ],
    verbose_mode=True,
    threshold=0.5,
    model=CustomGPTEnforced("gpt-3.5-turbo-0125"),
)

#########################################################
##### measure Metrics
#########################################################

# answer_relevancy.measure(test_case)
# answer_relevancy_non_confine.measure(test_case)
# answer_relevancy_confine.measure(test_case)

# bias.measure(test_case)
# bias_non_confine.measure(test_case)
# bias_confine.measure(test_case)

# contextual_precision.measure(test_case)
# contextual_precision_confine.measure(test_case)
# contextual_precision_non_confine.measure(test_case)

# contextual_recall.measure(test_case)
# contextual_recall_confine.measure(test_case)
# contextual_recall_non_confine.measure(test_case)

# contextual_relevancy.measure(test_case)
# contextual_relevancy_confine.measure(test_case)
# contextual_relevancy_non_confine.measure(test_case)

# faithfulness.measure(test_case)
# faithfulness_confine.measure(test_case)
# faithfulness_non_confine.measure(test_case)

# hallucination.measure(test_case)
# hallucination_confine.measure(test_case)
# hallucination_non_confine.measure(test_case)

# summarization.measure(test_case)
# summarization_confine.measure(test_case)
# summarization_non_confine.measure(test_case)

# toxicity.measure(test_case)
# toxicity_confine.measure(test_case)
# toxicity_non_confine.measure(test_case)

# g_eval.measure(test_case)
# g_eval_confine.measure(test_case)
# g_eval_non_confine.measure(test_case)


#########################################################
##### a_measure metrics
#########################################################

import asyncio


async def test_answer_relevancy():
    result = await answer_relevancy.a_measure(test_case)
    print("Answer Relevancy:", result)


async def test_answer_relevancy_non_confine():
    result = await answer_relevancy_non_confine.a_measure(test_case)
    print("Answer Relevancy Non Confine:", result)


async def test_answer_relevancy_confine():
    result = await answer_relevancy_confine.a_measure(test_case)
    print("Answer Relevancy Confine:", result)


async def test_bias():
    result = await bias.a_measure(test_case)
    print("Bias:", result)


async def test_bias_non_confine():
    result = await bias_non_confine.a_measure(test_case)
    print("Bias Non Confine:", result)


async def test_bias_confine():
    result = await bias_confine.a_measure(test_case)
    print("Bias Confine:", result)


async def test_contextual_recall():
    result = await contextual_recall.a_measure(test_case)
    print("Contextual Recall:", result)


async def test_contextual_recall_non_confine():
    result = await contextual_recall_non_confine.a_measure(test_case)
    print("Contextual Recall Non Confine:", result)


async def test_contextual_recall_confine():
    result = await contextual_recall_confine.a_measure(test_case)
    print("Contextual Recall Confine:", result)


async def test_contextual_relevancy():
    result = await contextual_relevancy.a_measure(test_case)
    print("Contextual Relevancy:", result)


async def test_contextual_relevancy_non_confine():
    result = await contextual_relevancy_non_confine.a_measure(test_case)
    print("Contextual Relevancy Non Confine:", result)


async def test_contextual_relevancy_confine():
    result = await contextual_relevancy_confine.a_measure(test_case)
    print("Contextual Relevancy Confine:", result)


async def test_faithfulness():
    result = await faithfulness.a_measure(test_case)
    print("Faithfulness:", result)


async def test_faithfulness_non_confine():
    result = await faithfulness_non_confine.a_measure(test_case)
    print("Faithfulness Non Confine:", result)


async def test_faithfulness_confine():
    result = await faithfulness_confine.a_measure(test_case)
    print("Faithfulness Confine:", result)


async def test_hallucination():
    result = await hallucination.a_measure(test_case)
    print("Hallucination:", result)


async def test_hallucination_non_confine():
    result = await hallucination_non_confine.a_measure(test_case)
    print("Hallucination Non Confine:", result)


async def test_hallucination_confine():
    result = await hallucination_confine.a_measure(test_case)
    print("Hallucination Confine:", result)


async def test_summarization():
    result = await summarization.a_measure(test_case)
    print("Summarization:", result)


async def test_summarization_non_confine():
    result = await summarization_non_confine.a_measure(test_case)
    print("Summarization Non Confine:", result)


async def test_summarization_confine():
    result = await summarization_confine.a_measure(test_case)
    print("Summarization Confine:", result)


async def test_toxicity():
    result = await toxicity.a_measure(test_case)
    print("Toxicity:", result)


async def test_toxicity_non_confine():
    result = await toxicity_non_confine.a_measure(test_case)
    print("Toxicity Non Confine:", result)


async def test_toxicity_confine():
    result = await toxicity_confine.a_measure(test_case)
    print("Toxicity Confine:", result)


async def test_g_eval():
    result = await g_eval.a_measure(test_case)
    print("G Eval:", result)


async def test_g_eval_non_confine():
    result = await g_eval_non_confine.a_measure(test_case)
    print("G Eval Non Confine:", result)


async def test_g_eval_confine():
    result = await g_eval_confine.a_measure(test_case)
    print("G Eval Confine:", result)


async def main():
    await asyncio.gather(
        test_answer_relevancy(),
        test_answer_relevancy_non_confine(),
        test_answer_relevancy_confine(),
        test_bias(),
        test_bias_non_confine(),
        test_bias_confine(),
        test_contextual_recall(),
        test_contextual_recall_non_confine(),
        test_contextual_recall_confine(),
        test_contextual_relevancy(),
        test_contextual_relevancy_non_confine(),
        test_contextual_relevancy_confine(),
        test_faithfulness(),
        test_faithfulness_non_confine(),
        test_faithfulness_confine(),
        test_hallucination(),
        test_hallucination_non_confine(),
        test_hallucination_confine(),
        test_summarization(),
        test_summarization_non_confine(),
        test_summarization_confine(),
        test_toxicity(),
        test_toxicity_non_confine(),
        test_toxicity_confine(),
        test_g_eval(),
        test_g_eval_non_confine(),
        test_g_eval_confine(),
    )


# Run the main function
asyncio.run(main())



from pydantic import BaseModel, Field
from typing import Optional, List, Union, Dict


class MetricMetadata(BaseModel):
    metric: str
    threshold: float
    success: bool
    score: Optional[float] = None
    reason: Optional[str] = None
    strict_mode: Optional[bool] = Field(False, alias="strictMode")
    evaluation_model: Optional[str] = Field(None, alias="evaluationModel")
    error: Optional[str] = None
    evaluation_cost: Union[float, None] = Field(None, alias="evaluationCost")
    verbose_logs: Optional[str] = Field(None, alias="verboseLogs")


class LLMApiTestCase(BaseModel):
    name: str
    input: str
    actual_output: str = Field(..., alias="actualOutput")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    context: Optional[list] = Field(None)
    retrieval_context: Optional[list] = Field(None, alias="retrievalContext")
    # make optional, not all test cases in a conversation will be evaluated
    success: Union[bool, None] = Field(None)
    # make optional, not all test cases in a conversation will be evaluated
    metrics_metadata: Union[List[MetricMetadata], None] = Field(
        None, alias="metricsMetadata"
    )
    # make optional, not all test cases in a conversation will be evaluated
    run_duration: Union[float, None] = Field(None, alias="runDuration")
    # make optional, not all test cases in a conversation will be evaluated
    evaluation_cost: Union[float, None] = Field(None, alias="evaluationCost")
    order: Union[int, None] = Field(None)
    # These should map 1 to 1 from golden
    additional_metadata: Optional[Dict] = Field(
        None, alias="additionalMetadata"
    )
    comments: Optional[str] = Field(None)
    traceStack: Optional[dict] = Field(None)

    def update(self, metric_metadata: MetricMetadata):
        if self.metrics_metadata is None:
            self.metrics_metadata = [metric_metadata]
        else:
            self.metrics_metadata.append(metric_metadata)

        if self.success is None:
            # self.success will be None when it is a message
            # in that case we will be setting success for the first time
            self.success = metric_metadata.success
        else:
            if metric_metadata.success is False:
                self.success = False

        evaluationCost = metric_metadata.evaluation_cost
        if evaluationCost is None:
            return

        if self.evaluation_cost is None:
            self.evaluation_cost = evaluationCost
        else:
            self.evaluation_cost += evaluationCost


class ConversationalApiTestCase(BaseModel):
    name: str
    success: bool
    # metrics_metadata can be None when we're not evaluating using conversational metrics
    metrics_metadata: Union[List[MetricMetadata], None] = Field(
        None, alias="metricsMetadata"
    )
    run_duration: float = Field(0.0, alias="runDuration")
    evaluation_cost: Union[float, None] = Field(None, alias="evaluationCost")
    messages: List[LLMApiTestCase] = Field(
        default_factory=lambda: [], alias="testCases"
    )
    order: Union[int, None] = Field(None)

    def update(self, metric_metadata: MetricMetadata, index: int):
        if index == -1:
            pass
        else:
            # if index != -1, update the metrics metadata of the specific message (llm test case)
            self.messages[index].update(metric_metadata)
            if metric_metadata.success is False:
                self.success = False

            evaluationCost = metric_metadata.evaluation_cost
            if evaluationCost is None:
                return

            if self.evaluation_cost is None:
                self.evaluation_cost = evaluationCost
            else:
                self.evaluation_cost += evaluationCost


class TestRunHttpResponse(BaseModel):
    testRunId: str
    projectId: str
    link: str



from typing import List
from pydantic import BaseModel, Field


class ContextualRecallVerdict(BaseModel):
    verdict: str
    reason: str = Field(default=None)


class Verdicts(BaseModel):
    verdicts: List[ContextualRecallVerdict]


class Reason(BaseModel):
    reason: str



from typing import Union, List, Optional, Any
import numpy as np
from deepeval.utils import normalize_text


# TODO: More scores are to be added
class Scorer:
    """This class calculates various Natural Language Processing (NLP) evaluation score.

    The scoring logic can be a simple algorithm or any statistical formula. There are some scores
    Which also uses an external model (BERTScore) in the scoring logic.
    """

    @classmethod
    def rouge_score(
        cls, target: str, prediction: str, score_type: str
    ) -> float:
        """Calculates the Rouge score for a given target and prediction.

        Rouge (Recall-Oriented Understudy for Gisting Evaluation) is a metric used for evaluating the quality of generated text,
        especially in tasks like text summarization.

        To utilize the rouge_score scoring method, be sure to `pip install rouge-score` before calling this method.

        Args:
            target (str): The actual label or target text.
            prediction (str): The generated text from the model or LLM.
            score_type (str): The Rouge score type (Options: 'rouge1', 'rouge2', 'rougeL').

        Returns:
            float: The Rouge score for the given target and prediction, based on the specified score type.
        """
        try:
            from rouge_score import rouge_scorer
        except:
            pass

        assert score_type in [
            "rouge1",
            "rouge2",
            "rougeL",
        ], "score_type can be either rouge1, rouge2 or rougeL"
        scorer = rouge_scorer.RougeScorer([score_type], use_stemmer=True)
        scores = scorer.score(target, prediction)
        return scores[score_type].fmeasure

    @classmethod
    def sentence_bleu_score(
        cls,
        references: Union[str, List[str]],
        prediction: str,
        bleu_type: Optional[str] = "bleu1",
    ) -> float:
        """Calculates the BLEU (Bilingual Evaluation Understudy) score for a given prediction compared to one or more reference sentences.

        BLEU is a metric used to evaluate the quality of machine-generated text by comparing it to one or more reference sentences.
        It measures the similarity of the generated text to the reference text based on n-grams.

        Args:
            references (Union[str, List[str]): A reference sentence or a list of reference sentences.
            prediction (str): The generated text or sentence to be evaluated.
            bleu_type (Optional[str]): The BLEU score type (Options: 'bleu1', 'bleu2', 'bleu3', 'bleu4'). Default is 'bleu1'.

        Returns:
            float: The BLEU score for the given prediction and references.
        """
        try:
            from nltk.tokenize import word_tokenize
            from nltk.translate.bleu_score import sentence_bleu
        except ModuleNotFoundError as e:
            print("Please install nltk module. Command: pip install nltk")

        assert bleu_type in [
            "bleu1",
            "bleu2",
            "bleu3",
            "bleu4",
        ], "Invalud bleu_type. Options: 'bleu1', 'bleu2', 'bleu3', 'bleu4'"
        targets = [references] if isinstance(references, str) else references
        tokenized_targets = [word_tokenize(target) for target in targets]
        tokenized_prediction = word_tokenize(prediction)
        bleu_weight_map = {
            "bleu1": (1, 0, 0, 0),
            "bleu2": (0, 1, 0, 0),
            "bleu3": (0, 0, 1, 0),
            "bleu4": (0, 0, 0, 1),
        }
        return sentence_bleu(
            tokenized_targets,
            tokenized_prediction,
            weights=bleu_weight_map[bleu_type],
        )

    @classmethod
    def exact_match_score(cls, target: str, prediction: str) -> int:
        """Metrics that calculates whether two sequences matches exactly or not.

        Args:
            target (str): The target string.
            prediction (str): The predicted string from the llm

        Returns:
            int: The exact match score.
        """
        if not prediction:
            return 0
        return 1 if prediction.strip() == target.strip() else 0

    @classmethod
    def quasi_exact_match_score(cls, target: str, prediction: str) -> int:
        if not prediction:
            return 0
        return 1 if normalize_text(target) == normalize_text(prediction) else 0

    # Todo: More mode based metrics to be added

    @classmethod
    def bert_score(
        cls,
        references: Union[str, List[str]],
        predictions: Union[str, List[str]],
        model: Optional[str] = "microsoft/deberta-large-mnli",
        lang: Optional[str] = "en",
    ) -> float:
        """
        Calculate BERTScore for one or more reference sentences compared to one or more prediction sentences using a specified BERT model.

        Args:
            references (Union[str, List[str]]): A single reference sentence or a list of reference sentences.
            predictions (Union[str, List[str]]): A single prediction sentence or a list of prediction sentences.
            model (Optional[str], optional): The name of the BERT model to be used for scoring. Defaults to "microsoft/deberta-large-mnli".
            lang (Optional[str], optional): The language code of the text, e.g., "en" for English. Defaults to "en".

        Returns:
            Dict[str, float]: A dictionary containing BERTScore metrics including precision, recall, and F1 score.
                - 'bert-precision' (float): BERTScore precision.
                - 'bert-recall' (float): BERTScore recall.
                - 'bert-f1' (float): BERTScore F1 score.

        Note:
            Before using this function, make sure to install the 'bert_score' module by running the following command:
            ```
            pip install bert-score
            ```
        """
        try:
            from bert_score import BERTScorer
        except ModuleNotFoundError as e:
            print(
                "Please install bert_score module. Command: pip install bert-score"
            )

        try:
            import torch
        except ModuleNotFoundError as e:
            print("Please install torch module. Command: pip install torch")

        # FIXME: Fix the case for mps
        device = "cuda" if torch.cuda.is_available() else "cpu"
        bert_scorer = BERTScorer(
            model_type=model,
            lang=lang,
            rescale_with_baseline=True,
            device=device,
        )

        if isinstance(predictions, str):
            predictions = [predictions]

        if isinstance(references, str):
            references = [references]

        if (
            isinstance(predictions, list)
            and isinstance(references, list)
            and not isinstance(references[0], list)
        ):
            if len(predictions) != len(references):
                references = [references]

        precision, recall, f1 = bert_scorer.score(
            cands=predictions, refs=references
        )
        return {
            "bert-precision": precision.detach().numpy().tolist(),
            "bert-recall": recall.detach().numpy().tolist(),
            "bert-f1": f1.detach().numpy().tolist(),
        }

    @classmethod
    def faithfulness_score(
        cls,
        target: str,
        prediction: str,
        model: Optional[str] = None,
        granularity: Optional[str] = None,
        device: Optional[str] = None,
    ) -> float:
        """Calculate the faithfulness score of a prediction compared to a target text using SummaCZS.

        This method computes a faithfulness score, which measures the extent to which a generated prediction matches the provided target text.
        The score is based on the SummaCZS (Summarization Competence with Zero-shot Supervision) model.

        Args:
            target (str): The reference target text for comparison.
            prediction (str): The generated prediction to be evaluated.
            model (Optional[str], optional): The SummaCZS model name to use. If not provided, the "vitc" model will be used by default.

        Returns:
            float: The computed faithfulness score. Higher values indicate greater faithfulness to the target text.

        Right now we are using score_one method under the hood. Instead of scoring multiple predictions for faithfullness.
        """
        try:
            from deepeval.models import SummaCModels
        except Exception as e:
            print(f"SummaCZS model can not be loaded.\n{e}")

        scorer = SummaCModels(
            model_name=model, granularity=granularity, device=device
        )
        return scorer(target, prediction)["score"]

    @classmethod
    def hallucination_score(
        cls, source: str, prediction: str, model: Optional[str] = None
    ) -> float:
        """Calculate the hallucination score of a prediction compared to a source text.

        This method computes a hallucination score, which measures the extent to which a generated prediction contains hallucinations.
        The score is based on the Vectara Hallucination Evaluation Model.

        Args:
            source (str): The source document where the information is summarized from.
            prediction (str): The generated summary that is validated against the source summary.

        Returns:
            float: The computed hallucination score. Lower values indicate greater hallucination.
        """
        try:
            from deepeval.models.hallucination_model import (
                HallucinationModel,
            )
        except ImportError as e:
            print(
                f"Vectera Hallucination detection model can not be loaded.\n{e}"
            )
        scorer = HallucinationModel(model_name=model)
        return scorer.model.predict([source, prediction])

    @classmethod
    def PII_score(
        cls, target: str, prediction: str, model: Optional[Any] = None
    ) -> float:
        raise NotImplementedError()

    @classmethod
    def neural_toxic_score(
        cls, prediction: str, model: Optional[str] = None
    ) -> Union[float, dict]:
        """
        Calculate the toxicity score of a given text prediction using the Detoxify model.

        Args:
            prediction (str): The text prediction to evaluate for toxicity.
            model (Optional[str], optional): The variant of the Detoxify model to use.
                Available variants: 'original', 'unbiased', 'multilingual'.
                If not provided, the 'original' variant is used by default.

        Returns:
            Union[float, dict]: The mean toxicity score, ranging from 0 (non-toxic) to 1 (highly toxic),
            and also a dictionary containing different types of toxicity score.

        For each model, we get mean toxicity score and a dictionary containing different toxicity score types.
        Examples:
        If model is 'original', we get the a dict with the following keys:
            - 'toxicity',
            - 'severe_toxicity',
            - 'obscene',
            - 'threat'
            - 'insult'
            - 'identity_attack'

        If model is 'unbiased', we get a dict with the same as keys as 'original', but
        along with `sexual_explicit`.

        If the model is 'multilingual', we get a dict same as the unbiasd one.
        """
        try:
            from deepeval.models.detoxify_model import DetoxifyModel
        except ImportError as e:
            print(f"Unable to import.\n {e}")
        scorer = DetoxifyModel(model_name=model)
        return scorer(prediction)

    @classmethod
    def answer_relevancy_score(
        cls,
        predictions: Union[str, List[str]],
        target: str,
        model_type: Optional[str] = None,
        model_name: Optional[str] = None,
    ) -> float:
        """Calculates the Answer relevancy score.

        Args:
            predictions (Union[str, List[str]]): The predictions from the model.
            target (str): The target on which we need to check relevancy.
            model_name (str): The type of the answer relevancy model. This can be either an self_encoder or a cross_encoder. By default it is cross_encoder.
            model_name (Optional[str], optional): The name of the model. Defaults to None.

        Returns:
            float: Answer relevancy score.
        """
        from sentence_transformers import util

        try:
            from deepeval.models import (
                AnswerRelevancyModel,
                CrossEncoderAnswerRelevancyModel,
            )
        except Exception as e:
            print(f"Unable to load AnswerRelevancyModel model.\n{e}")

        if model_type is not None:
            assert model_type in [
                "self_encoder",
                "cross_encoder",
            ], "model_type can be either 'self_encoder' or 'cross_encoder'"

        model_type = "cross_encoder" if model_type is None else model_type

        if model_type == "cross_encoder":
            assert isinstance(
                predictions, str
            ), "When model_type is 'cross_encoder', you can compare with one prediction and one target."
            answer_relevancy_model = CrossEncoderAnswerRelevancyModel(
                model_name=model_name
            )
            score = answer_relevancy_model(predictions, target)
        else:
            answer_relevancy_model = AnswerRelevancyModel(model_name=model_name)
            docs = (
                [predictions] if isinstance(predictions, str) else predictions
            )
            query_embedding = answer_relevancy_model(target)
            document_embedding = answer_relevancy_model(docs)
            scores = (
                util.dot_score(query_embedding, document_embedding)[0]
                .cpu()
                .tolist()
            )
            score = scores[0]
        return score

    @classmethod
    def neural_bias_score(cls, text: str, model: Optional[str] = None) -> float:
        try:
            from deepeval.models.unbias_model import UnBiasedModel
        except Exception as e:
            print(f"Unable to load UnBiasedModel.\n{e}")
        scorer = UnBiasedModel(model_name=model)
        return scorer(text)

    @classmethod
    def truth_identification_score(cls, target: str, prediction: str) -> int:
        """
        Metrics that calculates the number of correct true answers identified in the prediction.

        This method assumes both target and prediction are strings representing lists of integers,
        formatted like '1,2,3'. It converts these strings to lists of integers, counts how many items
        in the prediction list are also in the target list, and returns this count as the score.

        Args:
            target (str): The target string representing the list of correct answers.
            prediction (str): The predicted string from the LLM, representing the guessed answers.

        Returns:
            int: The number of correct answers identified.
        """
        try:
            if not prediction or not target:
                return 0  # Return score as 0 if prediction or target is empty

            # Convert strings to sorted lists of integers
            target_list = sorted(
                [int(item) for item in target.strip("[]").split(",") if item]
            )
            prediction_list = sorted(
                [
                    int(item)
                    for item in prediction.strip("[]").split(",")
                    if item
                ]
            )

            if not target_list:
                return 0  # Return 0 if target list is empty to avoid division by zero

            # Count the number of correct matches
            correct_matches = sum(
                1 for item in prediction_list if item in target_list
            )

            # Calculate percentage
            score_percentage = (correct_matches / len(target_list)) * 100

            return round(score_percentage)  # Return rounded percentage
        except Exception as e:
            return 0  # Return score as 0 in case of any exception

    def pass_at_k(self, n, c, k):
        """
        :param n: total number of samples
        :param c: number of correct samples
        :param k: k in pass@$k$
        """
        if n - c < k:
            return 1.0
        return 1.0 - np.prod(1.0 - k / np.arange(n - c + 1, n + 1))



from typing import Optional, Sequence, List, Union


def conform_contexts_type(
    contexts: Optional[Sequence[str]] = None,
) -> Union[List[str], None]:
    if contexts is None:
        return None

    return list(contexts)



import pytest
from deepeval import assert_test
from deepeval.metrics import (
    FaithfulnessMetric,
    AnswerRelevancyMetric,
    ContextualPrecisionMetric,
    ContextualRecallMetric,
    ContextualRelevancyMetric,
)
from deepeval.test_case import LLMTestCase

input = "Who won the FIFA World Cup in 2018 and what was the score?"
actual_output = (
    "Winners of the FIFA world cup were the French national football team"
)
expected_output = "French national football team"
retrieval_context = [
    "The FIFA World Cup in 2018 was won by the French national football team.",
    "I am birdy",
    "I am a froggy",
    "The French defeated Croatia 4-2 in the final FIFA match to claim the championship.",
]


@pytest.mark.skip(reason="openai is expensive")
def test_rag_metrics():
    test_case = LLMTestCase(
        input=input,
        actual_output=actual_output,
        retrieval_context=retrieval_context,
        expected_output=expected_output,
    )
    metric1 = AnswerRelevancyMetric(threshold=0.5)
    metric2 = FaithfulnessMetric(threshold=0.5)
    metric3 = ContextualRelevancyMetric(threshold=0.5)
    metric4 = ContextualPrecisionMetric(threshold=0.5)
    metric5 = ContextualRecallMetric(threshold=0.5)
    assert_test(test_case, [metric1, metric2, metric3, metric4, metric5])



from typing import List, Tuple, Dict, Union, Optional
import random
import asyncio

from deepeval.synthesizer.doc_chunker import (
    DocumentChunker,
    Chunk,
    get_embedding_similarity,
)
from deepeval.models.base_model import DeepEvalBaseEmbeddingModel
from deepeval.synthesizer.utils import initialize_embedding_model
from deepeval.utils import get_or_create_event_loop


class ContextGenerator:
    def __init__(
        self,
        document_paths: List[str],
        embedder: DeepEvalBaseEmbeddingModel,
        chunk_size: int = 1024,
        chunk_overlap: int = 0,
        async_mode: bool = False,
    ):
        self.embedder = embedder
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

        self.async_mode = async_mode
        self.document_paths: List[str] = document_paths

        # TODO: Potential bug, calling generate_goldens_from_docs
        # twice in a notebook enviornment will not refresh source_files_to_chunks_map
        self.source_files_to_chunks_map: Dict[str, List[Chunk]] = (
            self._load_docs()
        )

    ############### Generate Topics ########################
    def generate_contexts(
        self, num_context: int, max_context_size: int = 3
    ) -> Tuple[List[List[str]], List[str]]:
        contexts: List[List[str]] = []
        source_files: List[str] = []
        for path, document_chunks in self.source_files_to_chunks_map.items():
            num_chunks = len(document_chunks)
            num_context = min(num_context, num_chunks)
            clusters = self._get_n_random_clusters(
                path=path, n=num_context, cluster_size=max_context_size
            )
            for cluster in clusters:
                context = [chunk.content for chunk in cluster]
                contexts.append(context)
                source_files.append(cluster[0].source_file)

        return contexts, source_files

    ############### Load Docs #############################
    async def _a_load_docs(self) -> Dict[str, List[Chunk]]:
        async def a_process_document(path):
            doc_chunker = DocumentChunker(
                self.embedder, self.chunk_size, self.chunk_overlap
            )
            chunks = await doc_chunker.a_load_doc(path)
            return path, chunks

        source_files_to_chunks_map: Dict[str, List[Chunk]] = {}
        tasks = [a_process_document(path) for path in self.document_paths]
        path_to_chunks = await asyncio.gather(*tasks)
        for path, chunks in path_to_chunks:
            if path not in source_files_to_chunks_map:
                source_files_to_chunks_map[path] = []
            source_files_to_chunks_map[path].extend(chunks)
        return source_files_to_chunks_map

    def _load_docs(self) -> Dict[str, List[Chunk]]:
        if not self.async_mode:
            source_files_to_chunks_map: Dict[str, List[Chunk]] = {}
            for path in self.document_paths:
                doc_chunker = DocumentChunker(
                    self.embedder, self.chunk_size, self.chunk_overlap
                )
                chunks = doc_chunker.load_doc(path)
                if path not in source_files_to_chunks_map:
                    source_files_to_chunks_map[path] = []
                source_files_to_chunks_map[path].extend(chunks)
            return source_files_to_chunks_map
        else:
            loop = get_or_create_event_loop()
            return loop.run_until_complete(self._a_load_docs())

    ############### Search N Chunks ########################
    def _get_n_random_clusters(
        self, path: str, n: int, cluster_size: int
    ) -> List[List[Chunk]]:
        chunk_cluster = []
        random_chunks = self._get_n_random_chunks(path=path, n=n)
        for chunk in random_chunks:
            cluster = [chunk]
            cluster.extend(
                self._get_n_other_similar_chunks(
                    path=path, query_chunk=chunk, n=cluster_size - 1
                )
            )
            chunk_cluster.append(cluster)
        return chunk_cluster

    def _get_n_random_chunks(self, path: str, n: int) -> List[str]:
        document_chunks = self.source_files_to_chunks_map[path]
        n = min(len(document_chunks), n)
        return random.sample(document_chunks, n)

    def _get_n_other_similar_chunks(
        self,
        path: str,
        query_chunk: Chunk,
        n: int,
        threshold: float = 0.7,
    ) -> List[Chunk]:
        document_chunks = self.source_files_to_chunks_map[path]

        if not document_chunks:
            raise ValueError("Not enough chunks found for f{path}")

        # Confine n random chunks in case not enough chunks
        n = min(n, len(document_chunks))
        query_embedding = self.embedder.embed_text(query_chunk.content)
        similar_chunks = None

        similarities = []
        filtered_indices = []

        for i, chunk in enumerate(document_chunks):
            sim = get_embedding_similarity(query_embedding, (chunk.embedding))
            similarities.append(sim)

            if sim > threshold and document_chunks[i].id != query_chunk.id:
                filtered_indices.append(i)

        sorted_indices = sorted(
            filtered_indices, key=lambda i: similarities[i], reverse=True
        )
        top_n_indices = sorted_indices[:n]
        similar_chunks = [document_chunks[i] for i in top_n_indices]
        return similar_chunks



from .test_run import (
    TestRun,
    test_run_manager,
    TEMP_FILE_NAME,
    LLMApiTestCase,
    ConversationalApiTestCase,
)
from .hooks import on_test_run_end, invoke_test_run_end_hook
from .api import MetricMetadata
from .hyperparameters import log_hyperparameters



from .big_bench_hard.big_bench_hard import BigBenchHard
from .mmlu.mmlu import MMLU
from .hellaswag.hellaswag import HellaSwag
from .drop.drop import DROP
from .truthful_qa.truthful_qa import TruthfulQA
from .gsm8k.gsm8k import GSM8K
from .human_eval.human_eval import HumanEval



from pydantic import BaseModel, Field
from typing import List


class Opinions(BaseModel):
    opinions: List[str]


# ToxicMetric uses similar rubric to decoding trust: https://arxiv.org/abs/2306.11698
class ToxicityVerdict(BaseModel):
    verdict: str
    reason: str = Field(default=None)


class Verdicts(BaseModel):
    verdicts: List[ToxicityVerdict]


class Reason(BaseModel):
    reason: str



import re


class GSM8KTemplate:

    # Template was inspired by https://arxiv.org/pdf/2110.14168.pdf
    # Original method trained the generator on training set
    # Here we use the training set for COT few_shot prompting

    @staticmethod
    def generate_output(
        input: str, train_set: object, n_shots: int, enable_cot: bool
    ):
        prompt = ""

        # generate examples for n_shot inference
        if n_shots > 0:
            prompt = "The following are grade school math word problems\n\n"
        for i in range(n_shots):
            prompt += (
                GSM8KTemplate.format_example(train_set[i], enable_cot) + "\n\n"
            )

        # problem of interest
        prompt += "**Problem**: " + input + "\n**Answer**: \n\n"

        # define ouptut confinement
        prompt += (
            "Make sure to conclude your output with the numerical answer. "
        )
        if enable_cot:
            prompt += "Let's think step-by-step."
        else:
            prompt += "No explanation needed."

        return prompt

    @staticmethod
    def format_example(data: dict, enable_cot: bool):

        formatted_problem = ""
        question = data["question"]
        formatted_problem += "**Problem**: " + question + "\n"

        raw_answer = data["answer"]
        solution, answer = raw_answer.strip().split("\n#### ")
        if enable_cot:
            formatted_problem += "**Solution**: " + solution + "\n"
        formatted_problem += "**Answer**: " + answer

        return formatted_problem

    @staticmethod
    def format_answer(data: dict):
        raw_answer = data["answer"]
        answer = re.findall(r"#### (.*)", raw_answer)[0]
        return answer

    def format_subject(subject: str):
        return



class ContextualPrecisionTemplate:
    @staticmethod
    def generate_verdicts(input, expected_output, retrieval_context):
        return f"""Given the input, expected output, and retrieval context, please generate a list of JSON objects to determine whether each node in the retrieval context was remotely useful in arriving at the expected output.

**
IMPORTANT: Please make sure to only return in JSON format, with the 'verdicts' key as a list of JSON. These JSON only contain the `verdict` key that outputs only 'yes' or 'no', and a `reason` key to justify the verdict. In your reason, you should aim to quote parts of the context.
Example Retrieval Context: ["Einstein won the Nobel Prize for his discovery of the photoelectric effect", "He won the Nobel Prize in 1968.", "There was a cat."]
Example Input: "Who won the Nobel Prize in 1968 and for what?"
Example Expected Output: "Einstein won the Nobel Prize in 1968 for his discovery of the photoelectric effect."

Example:
{{
    "verdicts": [
        {{
            "verdict": "yes",
            "reason": "It clearly addresses the question by stating that 'Einstein won the Nobel Prize for his discovery of the photoelectric effect.'"
        }},
        {{
            "verdict": "yes",
            "reason": "The text verifies that the prize was indeed won in 1968."
        }},
        {{
            "verdict": "no",
            "reason": "'There was a cat' is not at all relevant to the topic of winning a Nobel Prize."
        }}
    ]  
}}
Since you are going to generate a verdict for each context, the number of 'verdicts' SHOULD BE STRICTLY EQUAL to that of the contexts.
**
        
Input:
{input}

Expected output:
{expected_output}

Retrieval Context:
{retrieval_context}

JSON:
"""

    @staticmethod
    def generate_reason(input, verdicts, score):
        # given the input and retrieval context for this input, where the verdict is whether ... and the node is the ..., give a reason for the score
        return f"""Given the input, retrieval contexts, and contextual precision score, provide a CONCISE summarize for the score. Explain why it is not higher, but also why it is at its current score.
The retrieval contexts is a list of JSON with three keys: `verdict`, `reason` (reason for the verdict) and `node`. `verdict` will be either 'yes' or 'no', which represents whether the corresponding 'node' in the retrieval context is relevant to the input. 
Contextual precision represents if the relevant nodes are ranked higher than irrelevant nodes. Also note that retrieval contexts is given IN THE ORDER OF THEIR RANKINGS.

**
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The score is <contextual_precision_score> because <your_reason>."
}}


DO NOT mention 'verdict' in your reason, but instead phrase it as irrelevant nodes. The term 'verdict' are just here for you to understand the broader scope of things.
Also DO NOT mention there are `reason` fields in the retrieval contexts you are presented with, instead just use the information in the `reason` field.
In your reason, you MUST USE the `reason`, QUOTES in the 'reason', and the node RANK (starting from 1, eg. first node) to explain why the 'no' verdicts should be ranked lower than the 'yes' verdicts.
When addressing nodes, make it explicit that it is nodes in retrieval context.
If the score is 1, keep it short and say something positive with an upbeat tone (but don't overdo it otherwise it gets annoying).
**

Contextual Precision Score:
{score}

Input:
{input}

Retrieval Contexts:
{verdicts}

JSON:
"""



from dataclasses import dataclass
from enum import Enum
from functools import wraps
from typing import Any, Callable, List, Union, Optional, Dict
from time import perf_counter
import traceback
from inspect import signature
import threading
from deepeval.utils import dataclass_to_dict


class TraceType(Enum):
    LLM = "LLM"
    RETRIEVER = "Retriever"
    EMBEDDING = "Embedding"
    TOOL = "Tool"
    AGENT = "Agent"
    AGENT_STEP = "Agent Step"
    CHAIN = "Chain"
    CHUNKING = "Chunking"
    NODE_PARSING = "Node Parsing"
    SYNTHESIZE = "Synthesize"
    QUERY = "Query"
    RERANKING = "Reranking"


class TraceStatus(Enum):
    SUCCESS = "Success"
    ERROR = "Error"


@dataclass
class LlmMetadata:
    model: Optional[str]
    tokenCount: Optional[Dict[str, int]] = None
    hyperparameters: Optional[Dict[str, Any]] = None
    outputMessages: Optional[List[Dict[str, str]]] = None
    llmPromptTemplate: Optional[Any] = None
    llmPromptTemplateVariables: Optional[Any] = None


@dataclass
class EmbeddingMetadata:
    model: Optional[str]


@dataclass
class BaseTrace:
    type: Union[TraceType, str]
    executionTime: float
    name: str
    input: dict
    output: dict
    status: TraceStatus
    traces: List["TraceData"]


@dataclass
class LlmTrace(BaseTrace):
    input: str
    llmMetadata: LlmMetadata


@dataclass
class EmbeddingTrace(BaseTrace):
    embeddingMetadata: EmbeddingMetadata


@dataclass
class GenericTrace(BaseTrace):
    type: str


TraceData = Union[LlmTrace, EmbeddingTrace, GenericTrace]


class TraceManager:
    def __init__(self):
        self._local = threading.local()

    def get_trace_stack(self):
        if not hasattr(self._local, "trace_stack"):
            self._local.trace_stack = []
            self._local.dict_trace_stack = None
        return self._local.trace_stack

    def clear_trace_stack(self):
        self.get_trace_stack().clear()

    def pop_trace_stack(self):
        if self.get_trace_stack():
            self.get_trace_stack().pop()

    def append_to_trace_stack(self, trace_instance):
        self.get_trace_stack().append(trace_instance)

    def set_dict_trace_stack(self, dict_trace_stack):
        self._local.dict_trace_stack = dict_trace_stack

    def get_and_reset_dict_trace_stack(self):
        dict_trace_stack = getattr(self._local, "dict_trace_stack", None)
        self._local.dict_trace_stack = None
        return dict_trace_stack


trace_manager = TraceManager()


def trace(
    type: str,
    name: Optional[str] = None,
    model: Optional[str] = None,
):
    assert isinstance(
        type, Union[TraceType, str]
    ), "'type' must be a 'TraceType' or str"

    if type in [TraceType.LLM, TraceType.EMBEDDING] and model is None:
        raise ValueError(f"{type} trace type requires a model.")
    assert model is None or isinstance(
        model, str
    ), "'model' must be a str or None"

    if type not in [TraceType.LLM, TraceType.EMBEDDING] and model is not None:
        raise ValueError(
            f"Parameter 'model' should not be provided for {type} trace types."
        )

    def decorator_trace(func: Callable):
        if type == TraceType.LLM:
            sig = signature(func)
            params = sig.parameters.values()

            # Check if it's an instance method, adjust parameter list if 'self' or 'cls' is present
            if any(p.name in ["self", "cls"] for p in params):
                params = [p for p in params if p.name not in ["self", "cls"]]

            # There should be exactly one parameter left of type list[str]
            if len(params) != 1:
                raise ValueError(
                    "Function of type `TraceType.LLM` must have exactly one parameter of type 'list[str]'"
                )

        @wraps(func)
        def wrapper(*args, **kwargs):
            sig = signature(func)
            if type == TraceType.LLM:
                input_str = (
                    args[1]
                    if "self" in sig.parameters or "cls" in sig.parameters
                    else args[0]
                )
                if not isinstance(input_str, str):
                    raise ValueError(
                        "Argument type for `TraceType.LLM` must be a string"
                    )

            bound_method = False
            # Check if it is called with 'self' or 'cls' parameter
            params = sig.parameters
            if args:
                first_param = next(iter(params))
                if first_param == "self" or first_param == "cls":
                    bound_method = True

            # Remove 'self' or 'cls' parameter if function is a method
            if bound_method:
                trace_args = args[1:]
            else:
                trace_args = args

            # Proceed to create your trace, using trace_args instead of args
            trace_instance_input = {"args": trace_args, "kwargs": kwargs}

            trace_instance = None
            effective_name = name if name is not None else func.__name__
            if type == TraceType.LLM:
                trace_instance = LlmTrace(
                    type=type,
                    executionTime=0,
                    name=effective_name,
                    input=input_str,
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                    llmMetadata=LlmMetadata(model=model),
                )
            elif type == TraceType.EMBEDDING:
                trace_instance = EmbeddingTrace(
                    type=type,
                    executionTime=0,
                    name=effective_name,
                    input=trace_instance_input,
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                    embeddingMetadata=EmbeddingMetadata(model=model),
                )
            else:
                trace_instance = GenericTrace(
                    type=type,
                    executionTime=0,
                    name=effective_name,
                    input=trace_instance_input,
                    output=None,
                    status=TraceStatus.SUCCESS,
                    traces=[],
                )

            trace_manager.append_to_trace_stack(trace_instance)
            start_time = perf_counter()
            try:
                result = func(*args, **kwargs)
                trace_instance.output = result

            except Exception as e:
                trace_instance.status = TraceStatus.ERROR
                trace_instance.output = {
                    "type": __builtins__["type"](e).__name__,
                    "message": str(e),
                    "traceback": traceback.format_exc(),
                }
                raise e

            finally:
                trace_instance.executionTime = perf_counter() - start_time

                current_trace_stack = trace_manager.get_trace_stack()
                if len(current_trace_stack) > 1:
                    parent_trace = current_trace_stack[-2]
                    parent_trace.traces.append(trace_instance)

                if len(current_trace_stack) == 1:
                    dict_representation = dataclass_to_dict(
                        current_trace_stack[0]
                    )
                    trace_manager.set_dict_trace_stack(dict_representation)
                    trace_manager.clear_trace_stack()
                else:
                    trace_manager.pop_trace_stack()

            return result

        return wrapper

    return decorator_trace


def set_token_usage(tokens: int):
    pass


def get_trace_stack():
    return trace_manager.get_and_reset_dict_trace_stack()



"""Tests for answer relevancy
"""

import pytest
from deepeval.metrics.base_metric import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.metrics import AnswerRelevancyMetric
from deepeval import assert_test

question = "What are the primary benefits of meditation?"
answer = """
Meditation offers a rich tapestry of benefits that touch upon various aspects of well-being. On a mental level, 
it greatly reduces stress and anxiety, fostering enhanced emotional health. This translates to better emotional 
regulation and a heightened sense of overall well-being. Interestingly, the practice of meditation has been around 
for centuries, evolving through various cultures and traditions, which underscores its timeless relevance.

Physically, it contributes to lowering blood pressure and alleviating chronic pain, which is pivotal for long-term health. 
Improved sleep quality is another significant benefit, aiding in overall physical restoration. Cognitively, meditation is a 
boon for enhancing attention span, improving memory, and slowing down age-related cognitive decline. Amidst these benefits, 
meditation's role in cultural and historical contexts is a fascinating side note, though not directly related to its health benefits.

Such a comprehensive set of advantages makes meditation a valuable practice for individuals seeking holistic improvement i
n both mental and physical health, transcending its historical and cultural origins.
"""

one = """
Meditation is an ancient practice, rooted in various cultural traditions, where individuals 
engage in mental exercises like mindfulness or concentration to promote mental clarity, emotional 
calmness, and physical relaxation. This practice can range from techniques focusing on breath, visual 
imagery, to movement-based forms like yoga. The goal is to bring about a sense of peace and self-awareness, 
enabling individuals to deal with everyday stress more effectively.
"""

two = """
One of the key benefits of meditation is its impact on mental health. It's widely used as a tool to 
reduce stress and anxiety. Meditation helps in managing emotions, leading to enhanced emotional health. 
It can improve symptoms of anxiety and depression, fostering a general sense of well-being. Regular practice 
is known to increase self-awareness, helping individuals understand their thoughts and emotions more clearly 
and reduce negative reactions to challenging situations.
"""

three = """
Meditation has shown positive effects on various aspects of physical health. It can lower blood pressure, 
reduce chronic pain, and improve sleep. From a cognitive perspective, meditation can sharpen the mind, increase 
attention span, and improve memory. It's particularly beneficial in slowing down age-related cognitive decline and 
enhancing brain functions related to concentration and attention.
"""


# Inherit BaseMetric
class FakeMetric(BaseMetric):
    def __init__(self, threshold: float = 0.5):
        self.threshold = threshold

    def measure(self, test_case: LLMTestCase):
        self.score = 1
        self.success = self.score >= self.threshold
        self.reason = "This metric looking good!"
        return self.score

    async def a_measure(self, test_case: LLMTestCase):
        self.score = 1
        self.success = self.score >= self.threshold
        self.reason = "This async metric looking good!"
        return self.score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Fake"


@pytest.mark.skip(reason="openai is very expensive")
def test_answer_relevancy():
    metric = FakeMetric(threshold=0.5)
    test_case = LLMTestCase(
        input="What is your name",
        actual_output="Idk",
        retrieval_context=[one, two, three],
    )
    assert_test(test_case, [metric])



import copy
import os
from typing import List, Optional, Union, Dict
import time
from dataclasses import dataclass

from deepeval.test_run.hyperparameters import process_hyperparameters
from deepeval.utils import (
    get_or_create_event_loop,
    set_verbose_mode,
    should_ignore_errors,
    should_use_cache,
    should_verbose_print,
)
from deepeval.telemetry import capture_evaluation_run
from deepeval.metrics import BaseMetric
from deepeval.metrics.indicator import (
    measure_metrics_with_indicator,
)
from deepeval.test_case import LLMTestCase, ConversationalTestCase
from deepeval.constants import PYTEST_RUN_TEST_NAME
from deepeval.test_run import (
    test_run_manager,
    LLMApiTestCase,
    ConversationalApiTestCase,
    MetricMetadata,
)
from deepeval.utils import (
    get_is_running_deepeval,
    set_indicator,
    capture_contextvars,
    update_contextvars,
)
from deepeval.test_run.cache import (
    test_run_cache_manager,
    Cache,
    CachedTestCase,
    CachedMetricData,
)
from deepeval.tracing import get_trace_stack


@dataclass
class TestResult:
    """Returned from run_test"""

    success: bool
    metrics_metadata: List[MetricMetadata]
    input: str
    actual_output: str
    expected_output: str
    context: List[str]
    retrieval_context: List[str]


def create_metric_metadata(metric: BaseMetric) -> MetricMetadata:
    if metric.error is not None:
        return MetricMetadata(
            metric=metric.__name__,
            threshold=metric.threshold,
            score=None,
            reason=None,
            success=False,
            strictMode=metric.strict_mode,
            evaluationModel=metric.evaluation_model,
            error=metric.error,
            evaluationCost=metric.evaluation_cost,
            verboseLogs=metric.verbose_logs,
        )
    else:
        return MetricMetadata(
            metric=metric.__name__,
            score=metric.score,
            threshold=metric.threshold,
            reason=metric.reason,
            success=metric.is_successful(),
            strictMode=metric.strict_mode,
            evaluationModel=metric.evaluation_model,
            error=None,
            evaluationCost=metric.evaluation_cost,
            verboseLogs=metric.verbose_logs,
        )


def create_test_result(
    test_case: Union[LLMApiTestCase, ConversationalApiTestCase],
) -> TestResult:
    if isinstance(test_case, ConversationalApiTestCase):
        tc = test_case.messages[len(test_case.messages) - 1]
    else:
        tc = test_case

    return TestResult(
        success=tc.success,
        metrics_metadata=tc.metrics_metadata,
        input=tc.input,
        actual_output=tc.actual_output,
        expected_output=tc.expected_output,
        context=tc.context,
        retrieval_context=tc.retrieval_context,
    )


def create_api_test_case(
    test_case: Union[LLMTestCase, ConversationalTestCase],
    index: Optional[int] = None,
    is_message: Optional[bool] = False,
    additional_metadata: Optional[Dict] = None,
    comments: Optional[str] = None,
) -> Union[LLMApiTestCase, ConversationalApiTestCase]:
    if isinstance(test_case, LLMTestCase):
        if is_message:
            success = None
            name = f"message_{index}"
            order = index

            # Manually set the metadata and comments on conversational test case
            # to each individual message (test case)
            test_case.additional_metadata = additional_metadata
            test_case.comments = comments
            trace_stack = None
        else:
            success = True
            name = os.getenv(PYTEST_RUN_TEST_NAME, f"test_case_{index}")
            order = test_case._dataset_rank
            trace_stack = get_trace_stack()

        return LLMApiTestCase(
            name=name,
            input=test_case.input,
            actualOutput=test_case.actual_output,
            expectedOutput=test_case.expected_output,
            context=test_case.context,
            retrievalContext=test_case.retrieval_context,
            success=success,
            metricsMetadata=None,
            runDuration=None,
            evaluationCost=None,
            order=order,
            additionalMetadata=test_case.additional_metadata,
            comments=test_case.comments,
            traceStack=trace_stack,
        )

    elif isinstance(test_case, ConversationalTestCase):
        return ConversationalApiTestCase(
            name=os.getenv(
                PYTEST_RUN_TEST_NAME, f"conversational_test_case_{index}"
            ),
            success=True,
            metricsMetadata=None,
            runDuration=0,
            evaluationCost=None,
            order=test_case._dataset_rank,
            testCases=[
                create_api_test_case(
                    tc,
                    i,
                    True,
                    test_case.additional_metadata,
                    test_case.comments,
                )
                for i, tc in enumerate(test_case.messages)
            ],
        )


def execute_test_cases(
    test_cases: List[Union[LLMTestCase, ConversationalTestCase]],
    metrics: List[BaseMetric],
    ignore_errors: bool,
    use_cache: bool,
    save_to_disk: bool = False,
    verbose_mode: Optional[bool] = None,
) -> List[TestResult]:
    test_results: List[TestResult] = []
    test_run_cache_manager.disable_write_cache = save_to_disk == False
    test_run_manager.save_to_disk = save_to_disk
    test_run = test_run_manager.get_test_run()

    if verbose_mode is not None:
        for metric in metrics:
            metric.verbose_mode = verbose_mode

    for index, test_case in enumerate(test_cases):
        with capture_evaluation_run("test case"):
            if isinstance(test_case, ConversationalTestCase):
                last_message_index = len(test_case.messages) - 1
            else:
                last_message_index = -1

            cached_test_case = None
            if use_cache and isinstance(
                test_case, LLMTestCase
            ):  # for now, don't use cache when it is a conversation
                cached_test_case = test_run_cache_manager.get_cached_test_case(
                    test_case, test_run.hyperparameters
                )

            ##### Metric Calculation #####
            # this can be a converational api or llm api test case
            api_test_case = create_api_test_case(test_case, index)
            new_cached_test_case: CachedTestCase = CachedTestCase()
            test_start_time = time.perf_counter()

            for metric in metrics:
                metric_metadata = None
                # cached_tet_case will always be false for conversationals
                if cached_test_case is not None:
                    cached_metric_data = Cache.get_metric_data(
                        metric, cached_test_case
                    )
                    if cached_metric_data:
                        metric_metadata = cached_metric_data.metric_metadata

                # metric_metadata will always be false for conversationals
                if metric_metadata is None:
                    metric.async_mode = False  # Override metric async
                    try:
                        # if conversational, manually extract the last one for now
                        # In the future, might add support for evaluating any message in a convo
                        if isinstance(test_case, ConversationalTestCase):
                            tc = test_case.messages[last_message_index]
                        else:
                            tc = test_case
                        metric.measure(tc)
                    except Exception as e:
                        if ignore_errors:
                            metric.error = str(e)  # Override metric error
                            metric.success = False  # Override metric success
                        else:
                            raise
                    metric_metadata = create_metric_metadata(metric)

                if isinstance(test_case, ConversationalTestCase):
                    # the index can be dynamic in the future, just not now
                    api_test_case.update(metric_metadata, last_message_index)
                else:
                    api_test_case.update(metric_metadata)

                if metric.error is None and isinstance(
                    test_case, LLMTestCase
                ):  # Only save to cache if metric didn't error and not conversational
                    cache_metric_metadata = create_metric_metadata(metric)
                    cache_metric_metadata.evaluation_cost = (
                        0  # Create copy and save 0 for cost
                    )
                    updated_cached_metric_data = CachedMetricData(
                        metric_metadata=cache_metric_metadata,
                        metric_configuration=Cache.create_metric_configuration(
                            metric
                        ),
                    )
                    new_cached_test_case.cached_metrics_data.append(
                        updated_cached_metric_data
                    )

            test_end_time = time.perf_counter()
            run_duration = test_end_time - test_start_time
            api_test_case.run_duration = run_duration

            ### Update Test Run ###
            test_run_manager.update_test_run(api_test_case, test_case)

            ### Cache Test Run ###
            if isinstance(
                test_case, LLMTestCase
            ):  # only cache if not conversational
                test_run_cache_manager.cache_test_case(
                    test_case,
                    new_cached_test_case,
                    test_run.hyperparameters,
                )
                test_run_cache_manager.cache_test_case(
                    test_case,
                    new_cached_test_case,
                    test_run.hyperparameters,
                    to_temp=True,
                )

            test_result = create_test_result(api_test_case)
            test_results.append(test_result)

    return test_results


async def a_execute_test_cases(
    test_cases: List[Union[LLMTestCase, ConversationalTestCase]],
    metrics: List[BaseMetric],
    ignore_errors: bool,
    use_cache: bool,
    save_to_disk: bool = False,
    verbose_mode: Optional[bool] = None,
) -> List[TestResult]:
    test_results: List[TestResult] = []
    test_run_cache_manager.disable_write_cache = save_to_disk == False
    test_run_manager.save_to_disk = save_to_disk
    test_run = test_run_manager.get_test_run()

    if verbose_mode is not None:
        for metric in metrics:
            metric.verbose_mode = verbose_mode

    for index, test_case in enumerate(test_cases):
        with capture_evaluation_run("test case"):
            cached_test_case = None
            # only use cache when NOT conversational test case
            if use_cache and isinstance(test_case, LLMTestCase):
                cached_test_case = test_run_cache_manager.get_cached_test_case(
                    test_case,
                    test_run.hyperparameters,
                )

            ##### Metric Calculation #####
            # api test case can be conversational
            api_test_case = create_api_test_case(test_case, index)

            new_cached_test_case: CachedTestCase = CachedTestCase()
            test_start_time = time.perf_counter()
            await measure_metrics_with_indicator(
                metrics, test_case, cached_test_case, ignore_errors
            )

            for metric in metrics:
                metric_metadata = create_metric_metadata(metric)
                if isinstance(test_case, ConversationalTestCase):
                    # index hardcoded as the last message for now
                    api_test_case.update(
                        metric_metadata, len(test_case.messages) - 1
                    )
                else:
                    api_test_case.update(metric_metadata)

                if metric.error is None and isinstance(
                    test_case, LLMTestCase
                ):  # Only save to cache if metric didn't error AND is not conversational
                    cache_metric_metadata = create_metric_metadata(metric)
                    cache_metric_metadata.evaluation_cost = (
                        0  # Create new copy and save 0 for cost
                    )
                    updated_cached_metric_data = CachedMetricData(
                        metric_metadata=cache_metric_metadata,
                        metric_configuration=Cache.create_metric_configuration(
                            metric
                        ),
                    )
                    new_cached_test_case.cached_metrics_data.append(
                        updated_cached_metric_data
                    )

            test_end_time = time.perf_counter()
            run_duration = test_end_time - test_start_time
            api_test_case.run_duration = run_duration

            ### Update Test Run ###
            test_run_manager.update_test_run(api_test_case, test_case)

            ### Cache Test Run ###
            if isinstance(
                test_case, LLMTestCase
            ):  # only cache if not conversational
                test_run_cache_manager.cache_test_case(
                    test_case,
                    new_cached_test_case,
                    test_run.hyperparameters,
                )
                test_run_cache_manager.cache_test_case(
                    test_case,
                    new_cached_test_case,
                    test_run.hyperparameters,
                    to_temp=True,
                )

            test_result = create_test_result(api_test_case)
            test_results.append(test_result)

    return test_results


def assert_test(
    test_case: Union[LLMTestCase, ConversationalTestCase],
    metrics: List[BaseMetric],
    run_async: bool = True,
):
    # TODO: keep this for now, blocking conversational metrics like KR
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    if run_async:
        loop = get_or_create_event_loop()
        test_result = loop.run_until_complete(
            a_execute_test_cases(
                [test_case],
                metrics,
                ignore_errors=should_ignore_errors(),
                use_cache=should_use_cache(),
                verbose_mode=should_verbose_print(),
                save_to_disk=get_is_running_deepeval(),
            )
        )[0]
    else:
        test_result = execute_test_cases(
            [test_case],
            metrics,
            ignore_errors=should_ignore_errors(),
            use_cache=should_use_cache(),
            verbose_mode=should_verbose_print(),
            save_to_disk=get_is_running_deepeval(),
        )[0]

    if not test_result.success:
        failed_metrics_metadata: List[MetricMetadata] = []
        # even for conversations, test_result right now is just the
        # result for the last message
        for metric_metadata in test_result.metrics_metadata:
            if metric_metadata.error is not None:
                failed_metrics_metadata.append(metric_metadata)
            else:
                # This try block is for user defined custom metrics,
                # which might not handle the score == undefined case elegantly
                try:
                    if not metric_metadata.success:
                        failed_metrics_metadata.append(metric_metadata)
                except:
                    failed_metrics_metadata.append(metric_metadata)

        failed_metrics_str = ", ".join(
            [
                f"{metrics_metadata.metric} (score: {metrics_metadata.score}, threshold: {metrics_metadata.threshold}, strict: {metrics_metadata.strict_mode}, error: {metrics_metadata.error})"
                for metrics_metadata in failed_metrics_metadata
            ]
        )
        raise AssertionError(f"Metrics: {failed_metrics_str} failed.")


def evaluate(
    test_cases: List[Union[LLMTestCase, ConversationalTestCase]],
    metrics: List[BaseMetric],
    hyperparameters: Optional[Dict[str, Union[str, int, float]]] = None,
    run_async: bool = True,
    show_indicator: bool = True,
    print_results: bool = True,
    write_cache: bool = True,
    use_cache: bool = False,
    ignore_errors: bool = False,
    verbose_mode: Optional[bool] = None,
):
    if hyperparameters is not None:
        if (
            hyperparameters.get("model") is None
            or hyperparameters.get("prompt template") is None
        ):
            raise ValueError(
                "A `model` and `prompt template` key must be provided when logging `hyperparameters`."
            )
        hyperparameters = process_hyperparameters(hyperparameters)

    # TODO: keep this for now, blocking conversational metrics like KR
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    set_indicator(show_indicator)

    test_run_manager.reset()
    start_time = time.perf_counter()
    if print_results:
        print("Evaluating test cases...")

    with capture_evaluation_run("evaluate()"):
        if run_async:
            loop = get_or_create_event_loop()
            test_results = loop.run_until_complete(
                a_execute_test_cases(
                    test_cases,
                    metrics,
                    ignore_errors=ignore_errors,
                    use_cache=use_cache,
                    verbose_mode=verbose_mode,
                    save_to_disk=write_cache,
                )
            )
        else:
            test_results = execute_test_cases(
                test_cases,
                metrics,
                ignore_errors=ignore_errors,
                use_cache=use_cache,
                verbose_mode=verbose_mode,
                save_to_disk=write_cache,
            )

    end_time = time.perf_counter()
    run_duration = end_time - start_time
    if print_results:
        for test_result in test_results:
            print_test_result(test_result)

        aggregate_metric_pass_rates(test_results)

    test_run = test_run_manager.get_test_run()
    test_run.hyperparameters = hyperparameters
    test_run_manager.save_test_run()
    test_run_manager.wrap_up_test_run(run_duration, display_table=False)
    return test_results


def print_test_result(test_result: TestResult):
    print("")
    print("=" * 70 + "\n")
    print("Metrics Summary\n")
    for metric_metadata in test_result.metrics_metadata:
        successful = True
        if metric_metadata.error is not None:
            successful = False
        else:
            # This try block is for user defined custom metrics,
            # which might not handle the score == undefined case elegantly
            try:
                if not metric_metadata.success:
                    successful = False
            except:
                successful = False

        if not successful:
            print(
                f"  - ❌ {metric_metadata.metric} (score: {metric_metadata.score}, threshold: {metric_metadata.threshold}, strict: {metric_metadata.strict_mode}, evaluation model: {metric_metadata.evaluation_model}, reason: {metric_metadata.reason}, error: {metric_metadata.error})"
            )
        else:
            print(
                f"  - ✅ {metric_metadata.metric} (score: {metric_metadata.score}, threshold: {metric_metadata.threshold}, strict: {metric_metadata.strict_mode}, evaluation model: {metric_metadata.evaluation_model}, reason: {metric_metadata.reason}, error: {metric_metadata.error})"
            )
        # if metrics_metadata.score_breakdown:
        #     for metric_name, score in metrics_metadata.score_breakdown.items():
        #         print(f"      - {metric_name} (score: {score})")

    print("")
    print("For test case:\n")
    print(f"  - input: {test_result.input}")
    print(f"  - actual output: {test_result.actual_output}")
    print(f"  - expected output: {test_result.expected_output}")
    print(f"  - context: {test_result.context}")
    print(f"  - retrieval context: {test_result.retrieval_context}")


def aggregate_metric_pass_rates(test_results: List[TestResult]) -> dict:
    metric_counts = {}
    metric_successes = {}

    for result in test_results:
        for metric_metadata in result.metrics_metadata:
            metric_name = metric_metadata.metric
            if metric_name not in metric_counts:
                metric_counts[metric_name] = 0
                metric_successes[metric_name] = 0
            metric_counts[metric_name] += 1
            if metric_metadata.success:
                metric_successes[metric_name] += 1

    metric_pass_rates = {
        metric: (metric_successes[metric] / metric_counts[metric])
        for metric in metric_counts
    }

    print("\n" + "=" * 70 + "\n")
    print("Overall Metric Pass Rates\n")
    for metric, pass_rate in metric_pass_rates.items():
        print(f"{metric}: {pass_rate:.2%} pass rate")
    print("\n" + "=" * 70 + "\n")

    return metric_pass_rates



from enum import Enum


class HumanEvalTask(Enum):
    HAS_CLOSE_ELEMENTS = "has_close_elements"
    SEPARATE_PAREN_GROUPS = "separate_paren_groups"
    TRUNCATE_NUMBER = "truncate_number"
    BELOW_ZERO = "below_zero"
    MEAN_ABSOLUTE_DEVIATION = "mean_absolute_deviation"
    INTERSPERSE = "intersperse"
    PARSE_NESTED_PARENS = "parse_nested_parens"
    FILTER_BY_SUBSTRING = "filter_by_substring"
    SUM_PRODUCT = "sum_product"
    ROLLING_MAX = "rolling_max"
    MAKE_PALINDROME = "make_palindrome"
    STRING_XOR = "string_xor"
    LONGEST = "longest"
    GREATEST_COMMON_DIVISOR = "greatest_common_divisor"
    ALL_PREFIXES = "all_prefixes"
    STRING_SEQUENCE = "string_sequence"
    COUNT_DISTINCT_CHARACTERS = "count_distinct_characters"
    PARSE_MUSIC = "parse_music"
    HOW_MANY_TIMES = "how_many_times"
    SORT_NUMBERS = "sort_numbers"
    FIND_CLOSEST_ELEMENTS = "find_closest_elements"
    RESCALE_TO_UNIT = "rescale_to_unit"
    FILTER_INTEGERS = "filter_integers"
    STRLEN = "strlen"
    LARGEST_DIVISOR = "largest_divisor"
    FACTORIZE = "factorize"
    REMOVE_DUPLICATES = "remove_duplicates"
    FLIP_CASE = "flip_case"
    CONCATENATE = "concatenate"
    FILTER_BY_PREFIX = "filter_by_prefix"
    GET_POSITIVE = "get_positive"
    IS_PRIME = "is_prime"
    FIND_ZERO = "find_zero"
    SORT_THIRD = "sort_third"
    UNIQUE = "unique"
    MAX_ELEMENT = "max_element"
    FIZZ_BUZZ = "fizz_buzz"
    SORT_EVEN = "sort_even"
    DECODE_CYCLIC = "decode_cyclic"
    PRIME_FIB = "prime_fib"
    TRIPLES_SUM_TO_ZERO = "triples_sum_to_zero"
    CAR_RACE_COLLISION = "car_race_collision"
    INCR_LIST = "incr_list"
    PAIRS_SUM_TO_ZERO = "pairs_sum_to_zero"
    CHANGE_BASE = "change_base"
    TRIANGLE_AREA = "triangle_area"
    FIB4 = "fib4"
    MEDIAN = "median"
    IS_PALINDROME = "is_palindrome"
    MODP = "modp"
    DECODE_SHIFT = "decode_shift"
    REMOVE_VOWELS = "remove_vowels"
    BELOW_THRESHOLD = "below_threshold"
    ADD = "add"
    SAME_CHARS = "same_chars"
    FIB = "fib"
    CORRECT_BRACKETING = "correct_bracketing"
    MONOTONIC = "monotonic"
    COMMON = "common"
    LARGEST_PRIME_FACTOR = "largest_prime_factor"
    SUM_TO_N = "sum_to_n"
    DERIVATIVE = "derivative"
    FIBFIB = "fibfib"
    VOWELS_COUNT = "vowels_count"
    CIRCULAR_SHIFT = "circular_shift"
    DIGITSUM = "digitSum"
    FRUIT_DISTRIBUTION = "fruit_distribution"
    PLUCK = "pluck"
    SEARCH = "search"
    STRANGE_SORT_LIST = "strange_sort_list"
    WILL_IT_FLY = "will_it_fly"
    SMALLEST_CHANGE = "smallest_change"
    TOTAL_MATCH = "total_match"
    IS_MULTIPLY_PRIME = "is_multiply_prime"
    IS_SIMPLE_POWER = "is_simple_power"
    IS_CUBE = "iscube"
    HEX_KEY = "hex_key"
    DECIMAL_TO_BINARY = "decimal_to_binary"
    IS_HAPPY = "is_happy"
    NUMERICAL_LETTER_GRADE = "numerical_letter_grade"
    PRIME_LENGTH = "prime_length"
    STARTS_ONE_ENDS = "starts_one_ends"
    SOLVE = "solve"
    ANTI_SHUFFLE = "anti_shuffle"
    GET_ROW = "get_row"
    SORT_ARRAY = "sort_array"
    ENCRYPT = "encrypt"
    NEXT_SMALLEST = "next_smallest"
    IS_BORED = "is_bored"
    ANY_INT = "any_int"
    ENCODE = "encode"
    SKJKASDKD = "skjkasdkd"
    CHECK_DICT_CASE = "check_dict_case"
    COUNT_UP_TO = "count_up_to"
    MULTIPLY = "multiply"
    COUNT_UPPER = "count_upper"
    CLOSEST_INTEGER = "closest_integer"
    MAKE_A_PILE = "make_a_pile"
    WORDS_STRING = "words_string"
    CHOOSE_NUM = "choose_num"
    ROUNDED_AVG = "rounded_avg"
    UNIQUE_DIGITS = "unique_digits"
    BY_LENGTH = "by_length"
    EVEN_ODD_PALINDROME = "even_odd_palindrome"
    COUNT_NUMS = "count_nums"
    MOVE_ONE_BALL = "move_one_ball"
    EXCHANGE = "exchange"
    HISTOGRAM = "histogram"
    REVERSE_DELETE = "reverse_delete"
    ODD_COUNT = "odd_count"
    MINSUBARRAYSUM = "minSubArraySum"
    MAX_FILL = "max_fill"
    SELECT_WORDS = "select_words"
    GET_CLOSEST_VOWEL = "get_closest_vowel"
    MATCH_PARENS = "match_parens"
    MAXIMUM = "maximum"
    SOLUTION = "solution"
    ADD_ELEMENTS = "add_elements"
    GET_ODD_COLLATZ = "get_odd_collatz"
    VALID_DATE = "valid_date"
    SPLIT_WORDS = "split_words"
    IS_SORTED = "is_sorted"
    INTERSECTION = "intersection"
    PROD_SIGNS = "prod_signs"
    MINPATH = "minPath"
    TRI = "tri"
    DIGITS = "digits"
    IS_NESTED = "is_nested"
    SUM_SQUARES = "sum_squares"
    CHECK_IF_LAST_CHAR_IS_A_LETTER = "check_if_last_char_is_a_letter"
    CAN_ARRANGE = "can_arrange"
    LARGEST_SMALLEST_INTEGERS = "largest_smallest_integers"
    COMPARE_ONE = "compare_one"
    IS_EQUAL_TO_SUM_EVEN = "is_equal_to_sum_even"
    SPECIAL_FACTORIAL = "special_factorial"
    FIX_SPACES = "fix_spaces"
    FILE_NAME_CHECK = "file_name_check"
    WORDS_IN_SENTENCE = "words_in_sentence"
    SIMPLIFY = "simplify"
    ORDER_BY_POINTS = "order_by_points"
    SPECIALFILTER = "specialFilter"
    GET_MAX_TRIPLES = "get_max_triples"
    BF = "bf"
    SORTED_LIST_SUM = "sorted_list_sum"
    X_OR_Y = "x_or_y"
    DOUBLE_THE_DIFFERENCE = "double_the_difference"
    COMPARE = "compare"
    STRONGEST_EXTENSION = "Strongest_Extension"
    CYCPATTERN_CHECK = "cycpattern_check"
    EVEN_ODD_COUNT = "even_odd_count"
    INT_TO_MINI_ROMAN = "int_to_mini_roman"
    RIGHT_ANGLE_TRIANGLE = "right_angle_triangle"
    FIND_MAX = "find_max"
    EAT = "eat"
    DO_ALGEBRA = "do_algebra"
    STRING_TO_MD5 = "string_to_md5"
    GENERATE_INTEGERS = "generate_integers"



from typing import Optional

from deepeval.api import Api, Endpoints
from deepeval.event.api import APIFeedback


def send_feedback(
    event_id: str,
    rating: int,
    expected_response: Optional[str] = None,
    explanation: Optional[str] = None,
    fail_silently: Optional[bool] = False,
    raise_expection: Optional[bool] = True,
) -> str:
    try:
        api_event = APIFeedback(
            eventId=event_id,
            rating=rating,
            expectedResponse=expected_response,
            explanation=explanation,
        )
        api = Api()
        try:
            body = api_event.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            # Pydantic version below 2.0
            body = api_event.dict(by_alias=True, exclude_none=True)

        api.post_request(
            endpoint=Endpoints.FEEDBACK_ENDPOINT.value,
            body=body,
        )

        return
    except Exception as e:
        if fail_silently:
            return

        if raise_expection:
            raise (e)
        else:
            print(str(e))



"""Tests for various benchmark functions.
"""

import pytest
from deepeval.benchmarks import GSM8K
from deepeval.benchmarks.gsm8k.template import GSM8KTemplate
from deepeval.dataset import Golden


@pytest.fixture
def golden():
    return Golden(
        input="Natalia sold clips to 48 of her friends in April, and then she sold half as many clips in May. How many clips did Natalia sell altogether in April and May?",
        expected_output="72",
    )


@pytest.fixture
def train_sample():
    return {
        "question": "Natalia sold clips to 48 of her friends in April, and then she sold half as many clips in May. How many clips did Natalia sell altogether in April and May?",
        "answer": "Natalia sold 48/2 = <<48/2=24>>24 clips in May.\nNatalia sold 48+24 = <<48+24=72>>72 clips altogether in April and May.\n#### 72",
    }


def test_gsm8k_template(golden, train_sample):
    max_n_shots = 3
    train_set = [train_sample for _ in range(max_n_shots)]

    no_cot_prompt = GSM8KTemplate.generate_output(
        golden.input, train_set, n_shots=max_n_shots, enable_cot=False
    )
    assert (
        "think step-by-step" not in no_cot_prompt
    ), f"Did not expect to see chain-of-thought in the prompt {no_cot_prompt}"

    cot_prompt = GSM8KTemplate.generate_output(
        golden.input, train_set, n_shots=max_n_shots, enable_cot=True
    )
    assert (
        "think step-by-step" in cot_prompt
    ), f"Expected to see chain-of-thought in the prompt {cot_prompt}"

    cot_prompt_zero_shot = GSM8KTemplate.generate_output(
        golden.input, train_set, n_shots=max_n_shots, enable_cot=True
    )
    assert (
        "think step-by-step" in cot_prompt_zero_shot
    ), f"Expected to see chain-of-thought in the prompt {cot_prompt_zero_shot}"



from typing import Optional, List
from langchain_openai import OpenAIEmbeddings, AzureOpenAIEmbeddings

from deepeval.key_handler import KeyValues, KEY_FILE_HANDLER
from deepeval.models import DeepEvalBaseLLM, DeepEvalBaseEmbeddingModel

valid_openai_embedding_models = [
    "text-embedding-3-small",
    "text-embedding-3-large",
    "text-embedding-ada-002",
]

default_openai_embedding_model = "text-embedding-3-small"


class OpenAIEmbeddingModel(DeepEvalBaseEmbeddingModel):
    def __init__(
        self,
        model: Optional[str] = None,
        *args,
        **kwargs,
    ):
        model_name = None
        if isinstance(model, str):
            model_name = model
            if model_name not in valid_openai_embedding_models:
                raise ValueError(
                    f"Invalid model. Available OpenAI Embedding models: {', '.join(model for model in valid_openai_embedding_models)}"
                )
        elif model is None:
            model_name = default_openai_embedding_model
        self.args = args
        self.kwargs = kwargs
        super().__init__(model_name)

    def load_model(self):
        print("!!!!!")
        if self.should_use_azure_openai():
            openai_api_key = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_API_KEY
            )

            openai_api_version = KEY_FILE_HANDLER.fetch_data(
                KeyValues.OPENAI_API_VERSION
            )
            azure_embedding_deployment = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_EMBEDDING_DEPLOYMENT_NAME
            )
            azure_endpoint = KEY_FILE_HANDLER.fetch_data(
                KeyValues.AZURE_OPENAI_ENDPOINT
            )

            return AzureOpenAIEmbeddings(
                openai_api_version=openai_api_version,
                azure_deployment=azure_embedding_deployment,
                azure_endpoint=azure_endpoint,
                openai_api_key=openai_api_key,
                **self.kwargs,
            )

        return OpenAIEmbeddings(model=self.model_name, **self.kwargs)

    def embed_text(self, text: str) -> List[float]:
        embedding_model = self.load_model()
        return embedding_model.embed_query(text)

    def embed_texts(self, texts: List[str]) -> List[List[float]]:
        embedding_model = self.load_model()
        return embedding_model.embed_documents(texts)

    async def a_embed_text(self, text: str) -> List[float]:
        embedding_model = self.load_model()
        return await embedding_model.aembed_query(text)

    async def a_embed_texts(self, texts: List[str]) -> List[List[float]]:
        embedding_model = self.load_model()
        return await embedding_model.aembed_documents(texts)

    def should_use_azure_openai(self):
        value = KEY_FILE_HANDLER.fetch_data(KeyValues.USE_AZURE_OPENAI)
        return value.lower() == "yes" if value is not None else False

    def get_model_name(self):
        if self.should_use_azure_openai():
            return "azure openai"
        elif self.model_name:
            return self.model_name



import logging
from typing import Any


def trace_langchain():
    try:
        from wrapt import wrap_function_wrapper
        from deepeval.integrations.langchain.callback import (
            LangChainCallbackHandler,
        )
        from deepeval.integrations.langchain import _BaseCallbackManagerInit

        wrap_function_wrapper(
            module="langchain_core.callbacks",
            name="BaseCallbackManager.__init__",
            wrapper=_BaseCallbackManagerInit(LangChainCallbackHandler()),
        )
        logging.info("Langchain tracing setup completed.")
    except Exception as e:
        logging.error(f"Error setting up Langchain tracing: {e}")


def trace_llama_index():
    try:
        from deepeval.integrations.llama_index.callback import (
            LlamaIndexCallbackHandler,
        )
        import llama_index.core

        llama_index.core.global_handler = LlamaIndexCallbackHandler()
        logging.info("Llama Index tracing setup completed.")
    except Exception as e:
        logging.error(f"Error setting up Llama Index tracing: {e}")



from typing import Optional

from deepeval.models import DeepEvalBaseLLM


def should_use_batch(model: DeepEvalBaseLLM, batch_size: Optional[int] = None):
    if batch_size is None:
        return False

    try:
        model.batch_generate([""])
    except Exception as AttributeError:
        return False
    except:
        raise

    return True



from typing import Union

from rich.live import Live
from rich.text import Text
from rich.table import Table
from rich.columns import Columns
from rich.console import Console
from rich.progress import Progress, BarColumn, SpinnerColumn, TextColumn


class RichManager:
    def __init__(self, show_table: bool, total_train_epochs: int) -> None:
        """
        Initialize RichManager.

        Args:
            show_table (bool): Flag to show or hide the table.
            total_train_epochs (int): Total number of training epochs.
        """
        self.show_table = show_table
        self.total_train_epochs = total_train_epochs
        self.console = Console()
        self.live = Live(auto_refresh=True, console=self.console)
        self.train_bar_started = False

        self.progress_bar_columns = [
            TextColumn(
                "{task.description} [progress.percentage][green][{task.percentage:>3.1f}%]:",
                justify="right",
            ),
            BarColumn(),
            TextColumn(
                "[green][ {task.completed}/{task.total} epochs ]",
                justify="right",
            ),
        ]
        self.spinner_columns = [
            TextColumn("{task.description}", justify="right"),
            SpinnerColumn(spinner_name="simpleDotsScrolling"),
        ]

        self.empty_column = Text("\n")

    def _initialize_progress_trackers(self) -> None:
        """
        Initialize progress trackers (progress and spinner columns).
        """
        self.progress = Progress(*self.progress_bar_columns, auto_refresh=False)
        self.spinner = Progress(*self.spinner_columns)

        self.progress_task = self.progress.add_task(
            "Train Progress", total=self.total_train_epochs
        )
        self.spinner_task = self.spinner.add_task("Initializing")

        column_list = [self.spinner, self.progress, self.empty_column]
        column_list.insert(0, Table()) if self.show_table else None

        column = Columns(column_list, equal=True, expand=True)
        self.live.update(column, refresh=True)

    def change_spinner_text(self, text: str) -> None:
        """
        Change the text displayed in the spinner.

        Args:
            text (str): Text to be displayed in the spinner.
        """
        self.spinner.reset(self.spinner_task, description=text)

    def stop(self) -> None:
        """Stop the live display."""
        self.live.stop()

    def start(self) -> None:
        """Start the live display and initialize progress trackers."""
        self.live.start()
        self._initialize_progress_trackers()

    def update(self, column: Columns) -> None:
        """
        Update the live display with a new column.

        Args:
            column (Columns): New column to be displayed.
        """
        self.live.update(column, refresh=True)

    def create_column(self) -> Union[Columns, Table]:
        """
        Create a new column with an optional table.

        Returns:
            Tuple[Columns, Table]: Tuple containing the new column and an optional table.
        """
        new_table = Table()

        column_list = [self.spinner, self.progress, self.empty_column]
        column_list.insert(0, new_table) if self.show_table else None

        column = Columns(column_list, equal=True, expand=True)
        return column, new_table

    def advance_progress(self) -> None:
        """Advance the progress tracker."""
        if not self.train_bar_started:
            self.progress.start()
            self.train_bar_started = True
        self.progress.update(self.progress_task, advance=1)
